import shutil
from datetime import datetime
from pathlib import Path
from typing import Any, Optional, TypeVar

import tomli
import tomli_w

from flood_adapt.dbs_classes.interface.database import IDatabase
from flood_adapt.dbs_classes.interface.element import AbstractDatabaseElement
from flood_adapt.misc.exceptions import (
    AlreadyExistsError,
    DatabaseError,
    DoesNotExistError,
    IsStandardObjectError,
    IsUsedInError,
)
from flood_adapt.objects.object_model import Object

T_OBJECTMODEL = TypeVar("T_OBJECTMODEL", bound=Object)


class DbsTemplate(AbstractDatabaseElement[T_OBJECTMODEL]):
    display_name: str
    dir_name: str
    _object_class: type[T_OBJECTMODEL]
    _higher_lvl_object: str

    def __init__(
        self, database: IDatabase, standard_objects: Optional[list[str]] = None
    ):
        """Initialize any necessary attributes."""
        self._database = database
        self.input_path = database.input_path / self.dir_name
        self.output_path = database.output_path / self.dir_name
        self.standard_objects = standard_objects

    def get(self, name: str) -> T_OBJECTMODEL:
        """Return an object of the type of the database with the given name.

        Parameters
        ----------
        name : str
            name of the object to be returned

        Returns
        -------
        Object
            object of the type of the specified object model

        Raises
        ------
        DoesNotExistError
            Raise error if the object does not exist.
        """
        # Make the full path to the object
        full_path = self.input_path / name / f"{name}.toml"

        # Check if the object exists
        if not Path(full_path).is_file():
            raise DoesNotExistError(name, self.display_name)

        # Load and return the object
        return self._object_class.load_file(full_path)

    def summarize_objects(self) -> dict[str, list[Any]]:
        """Return a dictionary with info on the objects that currently exist in the database.

        Returns
        -------
        dict[str, list[Any]]
            A dictionary that contains the keys: `name`, `description`, `path`  and `last_modification_date`.
            Each key has a list of the corresponding values, where the index of the values corresponds to the same object.
        """
        return self._get_object_summary()

    def copy(self, old_name: str, new_name: str, new_description: str):
        """Copy (duplicate) an existing object, and give it a new name.

        Parameters
        ----------
        old_name : str
            name of the existing measure
        new_name : str
            name of the new measure
        new_description : str
            description of the new measure

        Raises
        ------
        AlreadyExistsError
            Raise error if an object with the new name already exists.
        IsStandardObjectError
            Raise error if an object with the new name is a standard object.
        DatabaseError
            Raise error if the saving of the object fails.
        """
        copy_object = self.get(old_name)
        copy_object.name = new_name
        copy_object.description = new_description

        # After changing the name and description, re-trigger the validators
        copy_object.model_validate(copy_object)

        # Checking whether the new name is already in use
        self._validate_to_save(copy_object, overwrite=False)

        # Write only the toml file
        toml_path = self.input_path / new_name / f"{new_name}.toml"
        toml_path.parent.mkdir(parents=True)
        with open(toml_path, "wb") as f:
            tomli_w.dump(copy_object.model_dump(exclude_none=True), f)

        # Then copy all the accompanied files
        src = self.input_path / old_name
        dest = self.input_path / new_name

        EXCLUDE = [".spw", ".toml"]
        for file in src.glob("*"):
            if file.suffix in EXCLUDE:
                continue
            if file.is_dir():
                shutil.copytree(file, dest / file.name, dirs_exist_ok=True)
            else:
                shutil.copy2(file, dest / file.name)

    def save(
        self,
        object_model: T_OBJECTMODEL,
        overwrite: bool = False,
    ):
        """Save an object in the database and all associated files.

        This saves the toml file and any additional files attached to the object.

        Parameters
        ----------
        object_model : Object
            object to be saved in the database
        overwrite : bool, optional
            whether to overwrite the object if it already exists in the
            database, by default False

        Raises
        ------
        AlreadyExistsError
            Raise error if object to be saved already exists.
        IsStandardObjectError
            Raise error if object to be overwritten is a standard object.
        IsUsedInError
            Raise error if object to be overwritten is already in use.
        DatabaseError
            Raise error if the overwriting of the object fails.
        """
        self._validate_to_save(object_model, overwrite=overwrite)

        # If the folder doesnt exist yet, make the folder and save the object
        if not (self.input_path / object_model.name).exists():
            (self.input_path / object_model.name).mkdir()

        # Save the object and any additional files
        object_model.save(
            self.input_path / object_model.name / f"{object_model.name}.toml",
        )

    def delete(self, name: str, toml_only: bool = False):
        """Delete an already existing object as well as its outputs from the database.

        Parameters
        ----------
        name : str
            name of the object to be deleted
        toml_only : bool, optional
            whether to only delete the toml file or the entire folder. If the folder is empty after deleting the toml,
            it will always be deleted. By default False

        Raises
        ------
        IsStandardObjectError
            Raise error if object to be deleted is a standard object.
        IsUsedInError
            Raise error if object to be deleted is already in use.
        DoesNotExistError
            Raise error if object to be deleted does not exist.
        DatabaseError
            Raise error if the deletion of the object fails.
        """
        # Check if the object is a standard object. If it is, raise an error
        if self._check_standard_objects(name):
            raise IsStandardObjectError(name, self.display_name)

        # Check if object is used in a higher level object. If it is, raise an error
        if used_in := self.check_higher_level_usage(name):
            raise IsUsedInError(
                name, self.display_name, self._higher_lvl_object, used_in
            )

        # Check if the object exists
        toml_path = self.input_path / name / f"{name}.toml"
        if not toml_path.exists():
            raise DoesNotExistError(name, self.display_name)

        # Once all checks are passed, delete the object
        toml_path.unlink(missing_ok=True)

        # Delete the entire folder
        if not list(toml_path.parent.iterdir()):
            shutil.rmtree(toml_path.parent)
        elif not toml_only:
            try:
                shutil.rmtree(toml_path.parent)
            except OSError as e:
                raise DatabaseError(f"Failed to delete `{name}` due to: {e}") from e

        # Delete output
        if (self.output_path / name).exists():
            try:
                shutil.rmtree(self.output_path / name)
            except OSError as e:
                raise DatabaseError(
                    f"Failed to delete output of `{name}` due to: {e}"
                ) from e

    def _check_standard_objects(self, name: str) -> bool:
        """Check if an object is a standard object.

        Parameters
        ----------
        name : str
            name of the object to be checked

        Returns
        -------
        bool
            True if the object is a standard object, False otherwise
        """
        if self.standard_objects:
            return name in self.standard_objects
        return False

    def check_higher_level_usage(self, name: str) -> list[str]:
        """Check if an object is used in a higher level object.

        Parameters
        ----------
        name : str
            name of the object to be checked

        Returns
        -------
        list[str]
            list of higher level objects that use the object
        """
        # If this function is not implemented for the object type, it cannot be used in a higher
        # level object. By default, return an empty list
        return []

    def _get_object_summary(self) -> dict[str, list[Any]]:
        """Get a dictionary with all the toml paths and last modification dates that exist in the database of the given object_type.

        Returns
        -------
        dict[str, Any]
            A dictionary that contains the keys: `name`, `description`, `path`  and `last_modification_date`.
            Each key has a list of the corresponding values, where the index of the values corresponds to the same object.
        """
        # If the toml doesnt exist, we might be in the middle of saving a new object or could be a broken object.
        # In any case, we should not list it in the database
        directories = [
            dir
            for dir in self.input_path.iterdir()
            if (dir / f"{dir.name}.toml").is_file()
        ]
        paths = [Path(dir / f"{dir.name}.toml") for dir in directories]

        names = [self._read_variable_in_toml("name", path) for path in paths]
        descriptions = [
            self._read_variable_in_toml("description", path) for path in paths
        ]

        last_modification_date = [
            datetime.fromtimestamp(file.stat().st_mtime) for file in paths
        ]

        objects = {
            "name": names,
            "description": descriptions,
            "path": paths,
            "last_modification_date": last_modification_date,
        }
        return objects

    @staticmethod
    def _read_variable_in_toml(variable_name: str, toml_path: Path) -> str:
        with open(toml_path, "rb") as f:
            data = tomli.load(f)
        return data.get(variable_name, "")

    def _validate_to_save(self, object_model: T_OBJECTMODEL, overwrite: bool) -> None:
        """Validate if the object can be saved.

        Parameters
        ----------
        object_model : Object
            object to be validated

        Raises
        ------
        DatabaseError
            Raise error if name is already in use.
        """
        # Check if the object exists
        object_exists = object_model.name in self.summarize_objects()["name"]

        # If you want to overwrite the object, and the object already exists, first delete it. If it exists and you
        # don't want to overwrite, raise an error.
        if overwrite and object_exists:
            self.delete(object_model.name, toml_only=True)
        elif not overwrite and object_exists:
            raise AlreadyExistsError(object_model.name, self.display_name)



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Generic, TypeVar

from flood_adapt.objects.object_model import Object

T_OBJECT_MODEL = TypeVar("T_OBJECT_MODEL", bound=Object)


class AbstractDatabaseElement(ABC, Generic[T_OBJECT_MODEL]):
    input_path: Path
    output_path: Path

    @abstractmethod
    def __init__(self):
        """Abstract class for database elements."""
        pass

    @abstractmethod
    def get(self, name: str) -> T_OBJECT_MODEL:
        """Return the object of the type of the database with the given name.

        Parameters
        ----------
        name : str
            name of the object to be returned

        Returns
        -------
        IObject
            object of the type of the specified object model
        """
        pass

    @abstractmethod
    def summarize_objects(self) -> dict[str, list[Any]]:
        """Return a dictionary with info on the objects that currently exist in the database.

        Returns
        -------
        dict[str, Any]
            Includes 'name', 'description', 'path' and 'last_modification_date' info, as well as the objects themselves
        """
        pass

    @abstractmethod
    def copy(self, old_name: str, new_name: str, new_description: str):
        """Copy (duplicate) an existing object, and give it a new name.

        Parameters
        ----------
        old_name : str
            name of the existing measure
        new_name : str
            name of the new measure
        new_description : str
            description of the new measure
        """
        pass

    @abstractmethod
    def save(
        self,
        object_model: T_OBJECT_MODEL,
        overwrite: bool = False,
    ):
        """Save an object in the database.

        This only saves the toml file. If the object also contains a geojson file, this should be saved separately.

        Parameters
        ----------
        object_model : Object
            object to be saved in the database
        overwrite : bool, optional
            whether to overwrite the object if it already exists in the
            database, by default False
        toml_only : bool, optional
            whether to only save the toml file or all associated data. By default, save everything

        Raises
        ------
        ValueError
            Raise error if name is already in use.
        """
        pass

    @abstractmethod
    def delete(self, name: str, toml_only: bool = False):
        """Delete an already existing object in the database.

        Parameters
        ----------
        name : str
            name of the object to be deleted
        toml_only : bool, optional
            whether to only delete the toml file or the entire folder. If the folder is empty after deleting the toml,
            it will always be deleted. By default False

        Raises
        ------
        ValueError
            Raise error if object to be deleted is already in use.
        """
        pass

    @abstractmethod
    def check_higher_level_usage(self, name: str) -> list[str]:
        """Check if an object is used in a higher level object.

        Parameters
        ----------
        name : str
            name of the object to be checked

        Returns
        -------
        list[str]
            list of higher level objects that use the object
        """
        pass



from typing import Optional

from pydantic import BaseModel

from flood_adapt.objects.object_model import Object


class CurrentSituationModel(BaseModel):
    """
    The accepted input for a current situation in FloodAdapt.

    Attributes
    ----------
    projection : str
        The name of the projection. Should be a projection saved in the database.
    year : int
        The year of the current situation.
    """

    projection: str
    year: int


class Benefit(Object):
    """BaseModel describing the expected variables and data types of a Benefit analysis object.

    Attributes
    ----------
    name: str
        The name of the benefit analysis.
    description: str
        The description of the benefit analysis. Defaults to "".
    strategy : str
        The name of the strategy. Should be a strategy saved in the database.
    event_set : str
        The name of the event set. Should be an event set saved in the database.
    projection : str
        The name of the projection. Should be a projection saved in the database.
    future_year : int
        The future year for the analysis.
    current_situation : CurrentSituationModel
        The current situation model.
    baseline_strategy : str
        The name of the baseline strategy.
    discount_rate : float
        The discount rate for the analysis.
    implementation_cost : Optional[float]
        The implementation cost of the strategy. Defaults to None.
    annual_maint_cost : Optional[float]
        The annual maintenance cost of the strategy. Defaults to None.
    """

    strategy: str
    event_set: str
    projection: str
    future_year: int
    current_situation: CurrentSituationModel
    baseline_strategy: str
    discount_rate: float
    implementation_cost: Optional[float] = None
    annual_maint_cost: Optional[float] = None



from flood_adapt.dbs_classes.dbs_template import DbsTemplate
from flood_adapt.objects.projections.projections import Projection


class DbsProjection(DbsTemplate[Projection]):
    dir_name = "projections"
    display_name = "Projection"
    _object_class = Projection
    _higher_lvl_object = "Scenario"

    def check_higher_level_usage(self, name: str) -> list[str]:
        """Check if a projection is used in a scenario.

        Parameters
        ----------
        name : str
            name of the projection to be checked

        Returns
        -------
        list[str]
            list of scenarios that use the projection
        """
        # Get all the scenarios
        scenarios = [
            self._database.scenarios.get(scn)
            for scn in self._database.scenarios.summarize_objects()["name"]
        ]

        # Check if projection is used in a scenario
        used_in_scenario = [
            scenario.name for scenario in scenarios if name == scenario.projection
        ]

        return used_in_scenario



from flood_adapt import __version__
from flood_adapt.dbs_classes.interface.database import IDatabase
from flood_adapt.misc.log import FloodAdaptLogging
from flood_adapt.misc.utils import finished_file_exists, write_finished_file
from flood_adapt.objects.scenarios.scenarios import Scenario

logger = FloodAdaptLogging.getLogger("ScenarioRunner")


class ScenarioRunner:
    """class holding all information related to a scenario."""

    def __init__(self, database: IDatabase, scenario: Scenario) -> None:
        """Create a Scenario object."""
        self._database = database
        self._scenario = scenario
        self.site_info = self._database.site
        self.results_path = self._database.scenarios.output_path / self._scenario.name

        self._event = None
        self._projection = None
        self._strategy = None

        self._hazard_models = None
        self._impact_models = None

    @property
    def impact_models(self):
        """Return the list of impact models."""
        if self._impact_models is not None:
            return self._impact_models
        self._impact_models = self._database.static.get_impact_models()
        return self._impact_models

    @property
    def hazard_models(self):
        """Return the list of hazard models."""
        if self._hazard_models is not None:
            return self._hazard_models
        self._hazard_models = self._database.static.get_hazard_models()
        return self._hazard_models

    def _load_objects(self, scenario: Scenario) -> None:
        """Load objects from the database."""
        self._scenario = scenario
        self._event = self._database.events.get(scenario.event)
        self._projection = self._database.projections.get(scenario.projection)
        self._strategy = self._database.strategies.get(scenario.strategy)

    ### General methods ###
    def run(self) -> None:
        """Run hazard and impact models for the scenario."""
        self._database.has_run_hazard(self._scenario.name)
        self._load_objects(self._scenario)
        self.results_path.mkdir(parents=True, exist_ok=True)

        # Initiate the logger for all the integrator scripts.
        log_file = self.results_path.joinpath(f"logfile_{self._scenario.name}.log")
        with FloodAdaptLogging.to_file(file_path=log_file):
            logger.info(f"FloodAdapt version `{__version__}`")
            logger.info(f"Started evaluation of `{self._scenario.name}`")
            self._run_hazards()
            self._run_impacts()
            logger.info(f"Finished evaluation of `{self._scenario.name}`")

        # write finished file to indicate that the scenario has been run
        write_finished_file(self.results_path)

    def has_run_check(self):
        """Check if the scenario has been run."""
        return finished_file_exists(self.results_path)

    ### Hazard methods ###
    def _run_hazards(self) -> None:
        """Run the hazard model for the scenario."""
        if self.hazard_run_check():
            logger.info(f"Hazards for {self._scenario.name} have already been run.")
            return
        for model in self.hazard_models:
            model.run(self._scenario)

    def hazard_run_check(self) -> bool:
        """Check if the impact has been run.

        Returns
        -------
        bool
            _description_
        """
        return all(model.has_run(self._scenario) for model in self.hazard_models)

    ### Impact methods ###
    def _run_impacts(self) -> None:
        """Run the impact model(s)."""
        if self.impacts_run_check():
            logger.info(f"Impacts for {self._scenario.name} have already been run.")
            return
        for model in self.impact_models:
            model.run(self._scenario)

    def impacts_run_check(self) -> bool:
        """Check if the impact has been run.

        Returns
        -------
        bool
            _description_
        """
        return all(model.has_run(self._scenario) for model in self.impact_models)



import os
import shutil
from contextlib import contextmanager
from pathlib import Path
from typing import Union

from pydantic import BeforeValidator

from flood_adapt.misc.path_builder import (
    ObjectDir,
    db_path,
)


@contextmanager
def modified_environ(*remove, **update):
    """
    Temporarily updates the ``os.environ`` dictionary in-place.

    From https://github.com/laurent-laporte-pro/stackoverflow-q2059482/blob/master/demo/environ_ctx.py

    The ``os.environ`` dictionary is updated in-place so that the modification
    is sure to work in all situations.

    :param remove: Environment variables to remove.
    :param update: Dictionary of environment variables and values to add/update.
    """
    env = os.environ
    update = update or {}
    remove = remove or []

    # List of environment variables being updated or removed.
    stomped = (set(update.keys()) | set(remove)) & set(env.keys())
    # Environment variables and values to restore on exit.
    update_after = {k: env[k] for k in stomped}
    # Environment variables and values to remove on exit.
    remove_after = frozenset(k for k in update if k not in env)

    try:
        env.update(update)
        [env.pop(k, None) for k in remove]
        yield
    finally:
        env.update(update_after)
        [env.pop(k) for k in remove_after]


@contextmanager
def cd(newdir: Path):
    prevdir = Path().cwd()
    os.chdir(newdir)
    try:
        yield
    finally:
        os.chdir(prevdir)


def write_finished_file(path: Path):
    if not path.exists():
        path.mkdir(parents=True)
    with open(Path(path) / "finished.txt", "w") as f:
        f.write("run finished")


def finished_file_exists(path: Path):
    return (Path(path) / "finished.txt").exists()


def resolve_filepath(
    object_dir: ObjectDir, obj_name: str, path: Path | str | os.PathLike
) -> Path:
    """
    Determine whether a given path is an external file or a file in the database.

    Users can set the path to a file in the database directly, meaning it will be an absolute path.
    Users can also read the path from loading a toml file, meaning it will be a filename relative to the toml file.

    Parameters
    ----------
    object_dir : ObjectDir
        The directory name of the object in the database.
    obj_name : str
        The name of the object.
    path : Union[Path, str, os.PathLike]
        The path to the file, which can be an absolute path or a relative path.

    Returns
    -------
    Path
        The resolved path to the file.

    Raises
    ------
    FileNotFoundError
        If the file does not exist in either the provided path or the database path.
    """
    _path = Path(path)
    if str(_path) == _path.name:
        # this is a filename, so it is in the database
        src_path = db_path(object_dir=object_dir, obj_name=obj_name) / path
    else:
        # this is a path, so it is an external file
        src_path = Path(path)
    return src_path


def save_file_to_database(
    src_file: Path | str | os.PathLike, dst_dir: Path | str | os.PathLike
) -> Path:
    """Save a file to the database.

    Parameters
    ----------
    src_file : Path | str | os.PathLike
        Path to the file to be copied.
    dst_dir : Path | str | os.PathLike
        Path to the destination directory.

    Returns
    -------
    Path
        Path to the copied file

    Raises
    ------
    FileNotFoundError
        If the src_file does not exist at the given path
    """
    src_file = Path(src_file).resolve()
    dst_file = Path(dst_dir).resolve() / src_file.name

    if not src_file.exists():
        raise FileNotFoundError(
            f"Failed to find {src_file} when saving external file to the database as it does not exist."
        )
    if src_file == dst_file:
        return dst_file
    elif dst_file.exists():
        if dst_file.suffix == ".shp":
            for file in list(dst_file.parent.glob(f"{dst_file.stem}.*")):
                os.remove(file)
        else:
            os.remove(dst_file)

    dst_file.parent.mkdir(parents=True, exist_ok=True)
    if src_file.suffix == ".shp":
        for file in list(src_file.parent.glob(f"{src_file.stem}.*")):
            shutil.copy2(file, dst_file.parent.joinpath(file.name))
    else:
        shutil.copy2(src_file, dst_file)

    return dst_file


def copy_file_to_output_dir(file_path: Path, output_dir: Path) -> Path:
    output_dir = output_dir.resolve()
    if file_path == output_dir / file_path.name:
        return file_path
    output_dir.mkdir(parents=True, exist_ok=True)
    shutil.copy2(file_path, output_dir)
    return output_dir / file_path.name


def validate_file_extension(allowed_extensions: list[str]):
    """Validate the extension of the given path has one of the given suffixes.

    Examples
    --------
    >>> from pydantic import BaseModel
    >>> from pathlib import Path
    >>> from typing import Annotated

    >>> class MyClass(BaseModel):
    >>>     csv_path: Annotated[Path, validate_file_extension([".csv"])]
    """

    def _validator(value: Union[Path, str, os.PathLike]) -> Path:
        value = Path(value)
        if value.suffix not in allowed_extensions:
            raise ValueError(
                f"Invalid file extension: {value}. Allowed extensions are {', '.join(allowed_extensions)}."
            )
        return value

    return BeforeValidator(_validator)



import shutil
from typing import Any

import geopandas as gpd
import numpy as np
import numpy_financial as npf
import pandas as pd
import plotly.graph_objects as go
import tomli
import tomli_w
from fiat_toolbox.metrics_writer.fiat_read_metrics_file import MetricsFileReader

from flood_adapt.misc.exceptions import DatabaseError
from flood_adapt.misc.path_builder import (
    ObjectDir,
    TopLevelDir,
    db_path,
)
from flood_adapt.objects.benefits.benefits import Benefit
from flood_adapt.objects.scenarios.scenarios import Scenario


class BenefitRunner:
    """Object holding all attributes and methods related to a benefit analysis."""

    benefit: Benefit
    _scenarios: pd.DataFrame

    def __init__(self, database, benefit: Benefit):
        """Initialize function called when object is created through the load_file or load_dict methods."""
        self.database = database
        self.benefit = benefit

        # Get output path based on database path
        self.results_path = self.database.benefits.output_path.joinpath(
            self.benefit.name
        )
        self.site_info = self.database.site
        self.unit = self.site_info.fiat.config.damage_unit

    @property
    def scenarios(self) -> pd.DataFrame:
        """Get the scenarios of the benefit analysis.

        Returns
        -------
        pd.DataFrame
            a table with the scenarios of the Benefit analysis and their status
        """
        self._scenarios = self.check_scenarios()
        return self._scenarios

    @property
    def has_run(self):
        return self.has_run_check()

    @property
    def results(self):
        if hasattr(self, "_results"):
            return self._results
        self._results = self.get_output()
        return self._results

    def get_output(self) -> dict[str, Any]:
        if not self.has_run:
            raise RuntimeError(
                f"Cannot read output since benefit analysis '{self.benefit.name}' has not been run yet."
            )

        results_toml = self.results_path.joinpath("results.toml")
        results_html = self.results_path.joinpath("benefits.html")
        with open(results_toml, mode="rb") as fp:
            results = tomli.load(fp)
        results["html"] = str(results_html)
        self._results = results
        return results

    def has_run_check(self) -> bool:
        """Check if the benefit analysis has already been run.

        Returns
        -------
        has_run : bool
            True if the analysis has already been run, else False
        """
        # Output files to check
        results_toml = self.results_path.joinpath("results.toml")
        results_csv = self.results_path.joinpath("time_series.csv")
        results_html = self.results_path.joinpath("benefits.html")

        check = all(
            result.exists() for result in [results_toml, results_csv, results_html]
        )
        return check

    def check_scenarios(self) -> pd.DataFrame:
        """Check which scenarios are needed for this benefit calculation and if they have already been created.

        The scenarios attribute of the object is updated accordingly and the table of the scenarios is returned.

        Returns
        -------
        pd.DataFrame
            a table with the scenarios of the Benefit analysis and their status
        """
        # Define names of scenarios
        scenarios_calc = {
            "current_no_measures": {},
            "current_with_strategy": {},
            "future_no_measures": {},
            "future_with_strategy": {},
        }

        # Use the predefined names for the current projections and no measures strategy
        for scenario in scenarios_calc.keys():
            scenarios_calc[scenario]["event"] = self.benefit.event_set

            if "current" in scenario:
                scenarios_calc[scenario]["projection"] = (
                    self.benefit.current_situation.projection
                )
            elif "future" in scenario:
                scenarios_calc[scenario]["projection"] = self.benefit.projection

            if "no_measures" in scenario:
                scenarios_calc[scenario]["strategy"] = self.benefit.baseline_strategy
            else:
                scenarios_calc[scenario]["strategy"] = self.benefit.strategy

        # Get the available scenarios
        scenarios_avail = [
            self.database.scenarios.get(scn)
            for scn in self.database.scenarios.summarize_objects()["name"]
        ]

        # Check if any of the needed scenarios are already there
        for scenario in scenarios_calc.keys():
            scn_dict = scenarios_calc[scenario].copy()
            scn_dict["name"] = scenario
            scenario_obj = Scenario(**scn_dict)
            created = [
                scn_avl for scn_avl in scenarios_avail if scenario_obj == scn_avl
            ]
            if len(created) > 0:
                scenarios_calc[scenario]["scenario created"] = created[0].name
                scenarios_calc[scenario]["scenario run"] = (
                    self.database.scenarios.has_run_check(created[0].name)
                )
            else:
                scenarios_calc[scenario]["scenario created"] = "No"
                scenarios_calc[scenario]["scenario run"] = False

        df = pd.DataFrame(scenarios_calc).T
        scenarios = df.astype(
            dtype={
                "event": "str",
                "projection": "str",
                "strategy": "str",
                "scenario created": "str",
                "scenario run": bool,
            }
        )
        return scenarios

    def create_benefit_scenarios(self) -> None:
        """Create any scenarios that are needed for the (cost-)benefit assessment and are not there already.

        Parameters
        ----------
        benefit : Benefit
        """
        # Iterate through the scenarios needed and create them if not existing
        for _, row in self.scenarios.iterrows():
            if row["scenario created"] == "No":
                name = "_".join([row["projection"], row["event"], row["strategy"]])

                try:
                    self.database.scenarios.get(name)
                except DatabaseError:
                    # If the scenario does not exist, create it
                    scenario = Scenario(
                        name=name,
                        event=row["event"],
                        projection=row["projection"],
                        strategy=row["strategy"],
                    )
                    self.database.scenarios.save(scenario)

    def ready_to_run(self) -> bool:
        """Check if all the required scenarios have already been run.

        Returns
        -------
        bool
            True if required scenarios have been already run
        """
        check = all(self.scenarios["scenario run"])

        return check

    def run_cost_benefit(self):
        """Run the cost-benefit calculation for the total study area and the different aggregation levels."""
        # Throw an error if not all runs are finished
        if not self.ready_to_run():
            # First check is scenarios are there
            if "No" in self.scenarios["scenario created"].to_numpy():
                raise RuntimeError("Necessary scenarios have not been created yet.")
            scens = self.scenarios["scenario created"][~self.scenarios["scenario run"]]
            raise RuntimeError(
                f"Scenarios {', '.join(scens.values)} need to be run before the cost-benefit analysis can be performed"
            )

        # If path for results does not yet exist, make it, and if it does delete it and recreate it
        if not self.results_path.is_dir():
            self.results_path.mkdir(parents=True)
        else:
            shutil.rmtree(self.results_path)
            self.results_path.mkdir(parents=True)

        # Run the cost-benefit analysis
        self.cba()
        # Run aggregation benefits
        self.cba_aggregation()
        # Updates results
        self.has_run_check()

        # Cache results
        self.results

    def cba(self):
        """Cost-benefit analysis for the whole study area."""
        # Get EAD for each scenario and save to new dataframe
        scenarios = self.scenarios.copy(deep=True)
        scenarios["EAD"] = None

        scn_output_path = db_path(TopLevelDir.output, ObjectDir.scenario)

        # Get metrics per scenario
        for index, scenario in scenarios.iterrows():
            scn_name = scenario["scenario created"]
            collective_fn = scn_output_path.joinpath(
                scn_name, f"Infometrics_{scn_name}.csv"
            )
            collective_metrics = MetricsFileReader(
                collective_fn,
            ).read_metrics_from_file()

            # Fill scenarios EAD column with values from metrics
            scenarios.loc[index, "EAD"] = float(
                collective_metrics["Value"]["ExpectedAnnualDamages"]
            )

        # Get years of interest
        year_start = self.benefit.current_situation.year
        year_end = self.benefit.future_year

        # Calculate benefits
        cba = self._calc_benefits(
            years=[year_start, year_end],
            risk_no_measures=[
                scenarios.loc["current_no_measures", "EAD"],
                scenarios.loc["future_no_measures", "EAD"],
            ],
            risk_with_strategy=[
                scenarios.loc["current_with_strategy", "EAD"],
                scenarios.loc["future_with_strategy", "EAD"],
            ],
            discount_rate=self.benefit.discount_rate,
        )

        # Save indicators in dictionary
        results = {}
        # Get net present value of benefits
        results["benefits"] = cba["benefits_discounted"].sum()

        # Only if costs are provided do the full cost-benefit analysis
        cost_calc = (self.benefit.implementation_cost is not None) and (
            self.benefit.annual_maint_cost is not None
        )
        if cost_calc:
            cba = self._calc_costs(
                benefits=cba,
                implementation_cost=self.benefit.implementation_cost,
                annual_maint_cost=self.benefit.annual_maint_cost,
                discount_rate=self.benefit.discount_rate,
            )
            # Calculate costs
            results["costs"] = cba["costs_discounted"].sum()
            # Benefit to Cost Ratio
            results["BCR"] = np.round(results["benefits"] / results["costs"], 2)
            # Net present value
            results["NPV"] = cba["profits_discounted"].sum()
            # Internal Rate of Return
            results["IRR"] = np.round(npf.irr(cba["profits"]), 3)

        # Save results
        if not self.results_path.is_dir():
            self.results_path.mkdir(parents=True)

        # Save indicators in a toml file
        indicators = self.results_path.joinpath("results.toml")
        with open(indicators, "wb") as f:
            tomli_w.dump(results, f)

        # Save time-series in a csv
        time_series = self.results_path.joinpath("time_series.csv")
        cba.to_csv(time_series)

        # Make html
        self._make_html(cba)

    def cba_aggregation(self):
        """Zonal Benefits analysis for the different aggregation areas."""
        results_path = db_path(TopLevelDir.output, ObjectDir.scenario)
        # Get years of interest
        year_start = self.benefit.current_situation.year
        year_end = self.benefit.future_year

        # Get EAD for each scenario and save to new dataframe
        scenarios = self.scenarios.copy(deep=True)

        # Read in the names of the aggregation area types
        aggregations = [aggr.name for aggr in self.site_info.fiat.config.aggregation]

        # Check if equity information is available to define variables to use
        vars = []
        for i, aggr_name in enumerate(aggregations):
            if self.site_info.fiat.config.aggregation[i].equity is not None:
                vars.append(["EAD", "EWEAD"])
            else:
                vars.append(["EAD"])

        # Define which names are used in the metric tables
        var_metric = {"EAD": "ExpectedAnnualDamages", "EWEAD": "EWEAD"}

        # Prepare dictionary to save values
        risk = {}

        # Fill in the dictionary
        for i, aggr_name in enumerate(aggregations):
            risk[aggr_name] = {}
            values = {}
            for var in vars[i]:
                values[var] = []
            for index, scenario in scenarios.iterrows():
                scn_name = scenario["scenario created"]
                # Get available aggregation levels
                aggregation_fn = results_path.joinpath(
                    scn_name, f"Infometrics_{scn_name}_{aggr_name}.csv"
                )
                for var in vars[i]:
                    # Get metrics per scenario and per aggregation
                    aggregated_metrics = MetricsFileReader(
                        aggregation_fn,
                    ).read_aggregated_metric_from_file(var_metric[var])[2:]
                    aggregated_metrics = aggregated_metrics.loc[
                        aggregated_metrics.index.dropna()
                    ]
                    aggregated_metrics.name = scenario.name
                    values[var].append(aggregated_metrics)

            # Combine values in a single dataframe
            for var in vars[i]:
                risk[aggr_name][var] = pd.DataFrame(values[var]).T.astype(float)

        var_output = {"EAD": "Benefits", "EWEAD": "Equity Weighted Benefits"}

        # Calculate benefits
        benefits = {}
        for i, aggr_name in enumerate(aggregations):
            benefits[aggr_name] = pd.DataFrame()
            benefits[aggr_name].index = risk[aggr_name]["EAD"].index
            for var in vars[i]:
                for index, row in risk[aggr_name][var].iterrows():
                    cba = self._calc_benefits(
                        years=[year_start, year_end],
                        risk_no_measures=[
                            row["current_no_measures"],
                            row["future_no_measures"],
                        ],
                        risk_with_strategy=[
                            row["current_with_strategy"],
                            row["future_with_strategy"],
                        ],
                        discount_rate=self.benefit.discount_rate,
                    )
                    benefits[aggr_name].loc[row.name, var_output[var]] = cba[
                        "benefits_discounted"
                    ].sum()

        # Save results
        if not self.results_path.is_dir():
            self.results_path.mkdir(parents=True)

        # Save benefits per aggregation area (csv and gpkg)
        for i, aggr_name in enumerate(aggregations):
            csv_filename = self.results_path.joinpath(f"benefits_{aggr_name}.csv")
            benefits[aggr_name].to_csv(csv_filename, index=True)

            # Load aggregation areas
            ind = [
                i
                for i, n in enumerate(self.site_info.fiat.config.aggregation)
                if n.name == aggr_name
            ][0]
            aggr_areas_path = (
                db_path(TopLevelDir.static)
                / self.site_info.fiat.config.aggregation[ind].file
            )
            aggr_areas = gpd.read_file(aggr_areas_path, engine="pyogrio")
            # Define output path
            outpath = self.results_path.joinpath(f"benefits_{aggr_name}.gpkg")
            # Save file
            aggr_areas = aggr_areas.join(
                benefits[aggr_name],
                on=self.site_info.fiat.config.aggregation[ind].field_name,
            )
            aggr_areas.to_file(outpath, driver="GPKG")

    @staticmethod
    def _calc_benefits(
        years: list[int, int],
        risk_no_measures: list[float, float],
        risk_with_strategy: list[float, float],
        discount_rate: float,
    ) -> pd.DataFrame:
        """Calculate per year benefits and discounted benefits.

        Parameters
        ----------
        years : list[int, int]
            the current and future year for the analysis
        risk_no_measures : list[float, float]
            the current and future risk value without any measures
        risk_with_strategy : list[float, float]
            the current and future risk value with the strategy under investigation
        discount_rate : float
            the yearly discount rate used to calculated the total benefit

        Returns
        -------
        pd.DataFrame
            Dataframe containing the time-series of risks and benefits per year
        """
        benefits = pd.DataFrame(
            data={"risk_no_measures": np.nan, "risk_with_strategy": np.nan},
            index=np.arange(years[0], years[1] + 1),
        )
        benefits.index.names = ["year"]

        # Fill in dataframe
        for strat, risk in zip(
            ["no_measures", "with_strategy"], [risk_no_measures, risk_with_strategy]
        ):
            benefits.loc[years[0], f"risk_{strat}"] = risk[0]
            benefits.loc[years[1], f"risk_{strat}"] = risk[1]

        # Assume linear trend between current and future
        benefits = benefits.interpolate(method="linear")

        # Calculate benefits
        benefits["benefits"] = (
            benefits["risk_no_measures"] - benefits["risk_with_strategy"]
        )
        # Calculate discounted benefits using the provided discount rate
        benefits["benefits_discounted"] = benefits["benefits"] / (
            1 + discount_rate
        ) ** (benefits.index - benefits.index[0])

        return benefits

    @staticmethod
    def _calc_costs(
        benefits: pd.DataFrame,
        implementation_cost: float,
        annual_maint_cost: float,
        discount_rate: float,
    ) -> pd.DataFrame:
        """Calculate per year costs and discounted costs.

        Parameters
        ----------
        benefits : pd.DataFrame
            a time series of benefits per year (produced with __calc_benefits method)
        implementation_cost : float
            initial costs of implementing the adaptation strategy
        annual_maint_cost : float
            annual maintenance cost of the adaptation  strategy
        discount_rate : float
            yearly discount rate

        Returns
        -------
        pd.DataFrame
            Dataframe containing the time-series of benefits, costs and profits per year
        """
        benefits = benefits.copy()
        benefits["costs"] = np.nan
        # implementations costs at current year and maintenance from year 1
        benefits.loc[benefits.index[0], "costs"] = implementation_cost
        benefits.loc[benefits.index[1:], "costs"] = annual_maint_cost
        benefits["costs_discounted"] = benefits["costs"] / (1 + discount_rate) ** (
            benefits.index - benefits.index[0]
        )

        # Benefit to Cost Ratio
        benefits["profits"] = benefits["benefits"] - benefits["costs"]
        benefits["profits_discounted"] = benefits["profits"] / (1 + discount_rate) ** (
            benefits.index - benefits.index[0]
        )

        return benefits

    def _make_html(self, cba):
        """Make an html with the time-series of the benefits and discounted benefits."""
        # Save a plotly graph in an html
        fig = go.Figure()

        # Get only endpoints
        cba2 = cba.iloc[[0, -1]]

        # Add graph with benefits
        fig.add_trace(
            go.Scatter(
                x=cba.index,
                y=cba["benefits"],
                mode="lines",
                line_color="black",
                name="Interpolated benefits",
            )
        )
        fig.add_trace(
            go.Scatter(
                x=cba2.index,
                y=cba2["benefits"],
                mode="markers",
                marker_size=10,
                marker_color="rgba(53,217,44,1)",
                name="Calculated benefits",
            )
        )

        fig.add_trace(
            go.Scatter(
                x=cba.index,
                y=cba["benefits_discounted"],
                mode="lines",
                line_color="rgba(35,150,29,1)",
                name="Discounted benefits",
            ),
        )

        fig.add_trace(
            go.Scatter(
                x=cba.index,
                y=cba["benefits_discounted"],
                mode="none",
                fill="tozeroy",
                fillcolor="rgba(35,150,29,0.5)",
                name="Benefits",
            )
        )

        # Update xaxis properties
        fig.update_xaxes(title_text="Year")
        # Update yaxis properties
        fig.update_yaxes(title_text=f"Annual Benefits ({self.unit})")

        fig.update_layout(
            autosize=False,
            height=400,
            width=800,
            margin={"r": 0, "l": 0, "b": 0, "t": 0},
            font={"size": 12, "color": "black", "family": "Arial"},
        )

        # write html to results folder
        html = self.results_path.joinpath("benefits.html")
        fig.write_html(html)



from pathlib import Path
from typing import Optional

import geopandas as gpd
import numpy as np
import pandas as pd
import tomli
from pydantic import BaseModel, Field, model_validator

from flood_adapt.config.impacts import DamageType
from flood_adapt.objects.forcing import unit_system as us


class Layer(BaseModel):
    """
    Base class for layers in the GUI.

    Attributes
    ----------
    bins : list[float]
        The bins for the layer.
    colors : list[str]
        The colors for the layer.
    """

    bins: list[float]
    colors: list[str]

    @model_validator(mode="after")
    def check_bins_and_colors(self) -> "Layer":
        """Check that the bins and colors have the same length."""
        if (len(self.bins) + 1) != len(self.colors):
            raise ValueError(
                f"Number of bins ({len(self.bins)}) must be one less than number of colors ({len(self.colors)})"
            )
        return self


class FloodMapLayer(Layer):
    zbmax: float
    depth_min: float
    roads_min_zoom_level: int = 14


class AggregationDmgLayer(Layer):
    damage_decimals: Optional[int] = 0


class FootprintsDmgLayer(Layer):
    type: DamageType = DamageType.absolute
    damage_decimals: Optional[int] = 0
    buildings_min_zoom_level: int = 13


class BenefitsLayer(Layer):
    threshold: Optional[float] = None


class OutputLayers(BaseModel):
    """The configuration of the mapbox layers in the gui.

    Attributes
    ----------
    floodmap : FloodMapLayer
        The configuration of the floodmap layer.
    aggregation_dmg : AggregationDmgLayer
        The configuration of the aggregation damage layer.
    footprints_dmg : FootprintsDmgLayer
        The configuration of the footprints damage layer.

    benefits : BenefitsLayer
        The configuration of the benefits layer.
    """

    floodmap: FloodMapLayer
    aggregation_dmg: AggregationDmgLayer
    footprints_dmg: FootprintsDmgLayer

    benefits: Optional[BenefitsLayer] = None


class VisualizationLayer(Layer):
    """The configuration of a layer to visualize in the gui.

    name : str
        The name of the layer to visualize.
    long_name : str
        The long name of the layer to visualize.
    path : str
        The path to the layer data to visualize.
    field_name : str
        The field names of the layer to visualize.
    decimals : Optional[int]
        The number of decimals to use for the layer to visualize. default is None.
    """

    name: str
    long_name: str
    path: str
    field_name: str
    decimals: Optional[int] = None


_DEFAULT_BIN_NR = 4


def interpolate_hex_colors(
    start_hex="#FFFFFF", end_hex="#860000", number_bins=_DEFAULT_BIN_NR
):
    """
    Interpolate between two hex colors and returns a list of number_bins hex color codes.

    Parameters
    ----------
        start_hex : str
            Starting color in hex format (e.g., "#FFFFFF").
        end_hex : str
            Ending color in hex format (e.g., "#000000").
        number_bins : int
            Number of colors to generate between the start and end colors.

    Returns
    -------
        list[str]
            List of hex color codes interpolated between the start and end colors.
    """

    def hex_to_rgb(hex_color):
        hex_color = hex_color.lstrip("#")
        return tuple(int(hex_color[i : i + 2], 16) for i in (0, 2, 4))

    def rgb_to_hex(rgb_color):
        return "#{:02X}{:02X}{:02X}".format(*rgb_color)

    start_rgb = hex_to_rgb(start_hex)
    end_rgb = hex_to_rgb(end_hex)

    interpolated_colors = []
    for i in range(number_bins):
        ratio = i / (number_bins - 1) if number_bins > 1 else 0
        interpolated_rgb = tuple(
            int(start + (end - start) * ratio) for start, end in zip(start_rgb, end_rgb)
        )
        interpolated_colors.append(rgb_to_hex(interpolated_rgb))

    return interpolated_colors


class VisualizationLayers(BaseModel):
    """The configuration of the layers you might want to visualize in the gui.

    Attributes
    ----------
    default : Layer
        The default layer settings the visualization layers.
    layers : list[VisualizationLayer]
        The layers to visualize.
    """

    layers: list[VisualizationLayer] = Field(default_factory=list)

    def add_layer(
        self,
        name: str,
        long_name: str,
        path: str,
        field_name: str,
        database_path: Path,
        decimals: Optional[int] = None,
        bins: Optional[list[float]] = None,
        colors: Optional[list[str]] = None,
    ) -> None:
        if not Path(path).is_absolute():
            raise ValueError(f"Path {path} must be absolute.")

        data = gpd.read_file(path)
        if field_name not in data.columns:
            raise ValueError(
                f"Field name {field_name} not found in data. Available fields: {data.columns.tolist()}"
            )

        if bins is None:
            _, _bins = pd.qcut(
                data[field_name], _DEFAULT_BIN_NR, retbins=True, duplicates="drop"
            )
            bins = _bins.tolist()[1:-1]

        if decimals is None:
            non_zero_bins = [abs(b) for b in bins if b != 0]
            min_non_zero = min(non_zero_bins) if non_zero_bins else 1
            decimals = max(int(-np.floor(np.log10(min_non_zero))), 0)

        if colors is None:
            nr_bins = len(bins) + 1
            colors = interpolate_hex_colors(number_bins=nr_bins)

        relative_path = Path(path).relative_to(database_path / "static")
        self.layers.append(
            VisualizationLayer(
                bins=bins,
                colors=colors,
                name=name,
                long_name=long_name,
                path=relative_path.as_posix(),
                field_name=field_name,
                decimals=decimals,
            )
        )


class GuiUnitModel(BaseModel):
    """The unit system used in the GUI.

    Attributes
    ----------
    default_length_units : us.UnitTypesLength
        The length units used in the GUI.
    default_distance_units : us.UnitTypesLength
        The distance units used in the GUI.
    default_area_units : us.UnitTypesArea
        The area units used in the GUI.
    default_volume_units : us.UnitTypesVolume
        The volume units used in the GUI.
    default_velocity_units : us.UnitTypesVelocity
        The velocity units used in the GUI.
    default_direction_units : us.UnitTypesDirection
        The direction units used in the GUI.
    default_discharge_units : us.UnitTypesDischarge
        The discharge units used in the GUI.
    default_intensity_units : us.UnitTypesIntensity
        The intensity units used in the GUI.
    default_cumulative_units : us.UnitTypesLength
        The cumulative units used in the GUI.
    """

    default_length_units: us.UnitTypesLength
    default_distance_units: us.UnitTypesLength
    default_area_units: us.UnitTypesArea
    default_volume_units: us.UnitTypesVolume
    default_velocity_units: us.UnitTypesVelocity
    default_direction_units: us.UnitTypesDirection
    default_discharge_units: us.UnitTypesDischarge
    default_intensity_units: us.UnitTypesIntensity
    default_cumulative_units: us.UnitTypesLength

    @staticmethod
    def imperial() -> "GuiUnitModel":
        return GuiUnitModel(
            default_length_units=us.UnitTypesLength.feet,
            default_distance_units=us.UnitTypesLength.miles,
            default_area_units=us.UnitTypesArea.sf,
            default_volume_units=us.UnitTypesVolume.cf,
            default_velocity_units=us.UnitTypesVelocity.mph,
            default_direction_units=us.UnitTypesDirection.degrees,
            default_discharge_units=us.UnitTypesDischarge.cfs,
            default_intensity_units=us.UnitTypesIntensity.inch_hr,
            default_cumulative_units=us.UnitTypesLength.inch,
        )

    @staticmethod
    def metric() -> "GuiUnitModel":
        return GuiUnitModel(
            default_length_units=us.UnitTypesLength.meters,
            default_distance_units=us.UnitTypesLength.meters,
            default_area_units=us.UnitTypesArea.m2,
            default_volume_units=us.UnitTypesVolume.m3,
            default_velocity_units=us.UnitTypesVelocity.mps,
            default_direction_units=us.UnitTypesDirection.degrees,
            default_discharge_units=us.UnitTypesDischarge.cms,
            default_intensity_units=us.UnitTypesIntensity.mm_hr,
            default_cumulative_units=us.UnitTypesLength.millimeters,
        )


class SyntheticTideModel(BaseModel):
    """Configuration for the synthetic tide.

    Parameters
    ----------
    harmonic_amplitude : us.UnitfulLength
        The amplitude of the tide harmonic relative to the datum.
    datum : str
        The datum to which the harmonic amplitude is relative.
        Should be a datum defined in `site.sfincs.waterlevels.datums`
    """

    harmonic_amplitude: us.UnitfulLength
    datum: str


class PlottingModel(BaseModel):
    """
    The configuration of the plotting in the gui.

    Attributes
    ----------
    excluded_datums : list[str]
        A list of datums that will be excluded from the forcing plot in event windows.
    synthetic_tide : SyntheticTideModel
        The configuration of the synthetic tide.
    """

    synthetic_tide: SyntheticTideModel
    excluded_datums: list[str] = Field(default_factory=list)


class GuiModel(BaseModel):
    """The accepted input for the variable gui in Site.

    Attributes
    ----------
    units : GuiUnitModel
        The unit system used in the GUI.
    output_layers : OutputLayers
        The configuration of the mapbox layers in the GUI.
    visualization_layers : VisualizationLayers
        The configuration of the visualization layers in the GUI.
    plotting : PlottingModel
        The configuration for creating hazard forcing plots.
    """

    units: GuiUnitModel
    output_layers: OutputLayers
    visualization_layers: VisualizationLayers
    plotting: PlottingModel

    @staticmethod
    def read_toml(path: Path) -> "GuiModel":
        with open(path, mode="rb") as fp:
            toml_contents = tomli.load(fp)

        return GuiModel(**toml_contents)



import datetime
import gc
import logging
import math
import os
import re
import shutil
import warnings
from enum import Enum
from pathlib import Path
from typing import Optional, Union
from urllib.request import urlretrieve

import cht_observations.observation_stations as obs
import geopandas as gpd
import numpy as np
import pandas as pd
import rioxarray as rxr
import tomli
import tomli_w
import xarray as xr
from hydromt_fiat import FiatModel as HydromtFiatModel
from hydromt_fiat.data_apis.open_street_maps import get_buildings_from_osm
from hydromt_sfincs import SfincsModel as HydromtSfincsModel
from pydantic import BaseModel, Field
from shapely import MultiLineString, MultiPolygon, Polygon

from flood_adapt.adapter.fiat_adapter import _FIAT_COLUMNS
from flood_adapt.config.fiat import (
    FiatConfigModel,
    FiatModel,
)
from flood_adapt.config.gui import (
    AggregationDmgLayer,
    BenefitsLayer,
    FloodMapLayer,
    FootprintsDmgLayer,
    GuiModel,
    GuiUnitModel,
    OutputLayers,
    PlottingModel,
    SyntheticTideModel,
    VisualizationLayers,
)
from flood_adapt.config.hazard import (
    Cstype,
    CycloneTrackDatabaseModel,
    DatumModel,
    DemModel,
    FloodModel,
    ObsPointModel,
    RiverModel,
    SCSModel,
    SlrScenariosModel,
    WaterlevelReferenceModel,
)
from flood_adapt.config.impacts import (
    AggregationModel,
    BenefitsModel,
    BFEModel,
    EquityModel,
    FloodmapType,
    RiskModel,
    SVIModel,
)
from flood_adapt.config.sfincs import (
    SfincsConfigModel,
    SfincsModel,
)
from flood_adapt.config.site import (
    Site,
    StandardObjectModel,
)
from flood_adapt.dbs_classes.database import Database
from flood_adapt.misc.debug_timer import debug_timer
from flood_adapt.misc.log import FloodAdaptLogging
from flood_adapt.misc.utils import modified_environ
from flood_adapt.objects.events.event_set import EventSet
from flood_adapt.objects.forcing import unit_system as us
from flood_adapt.objects.forcing.tide_gauge import (
    TideGauge,
    TideGaugeSource,
)
from flood_adapt.objects.projections.projections import (
    PhysicalProjection,
    Projection,
    SocioEconomicChange,
)
from flood_adapt.objects.strategies.strategies import Strategy

logger = FloodAdaptLogging.getLogger("DatabaseBuilder")


def path_check(str_path: str, config_path: Optional[Path] = None) -> str:
    """Check if the given path is absolute and return the absolute path.

    Parameters
    ----------
    str_path : str
        The path to be checked.
    config_path : Optional[Path], default None
        The base path to resolve relative paths.

    Returns
    -------
    str
        The absolute path as a string.

    Raises
    ------
    ValueError
        If the path is not absolute and no config_path is provided.
    """
    path = Path(str_path)
    if not path.is_absolute():
        if config_path is not None:
            path = Path(config_path).parent.joinpath(path).resolve()
        else:
            raise ValueError(f"Value '{path}' should be an absolute path.")
    return path.as_posix()


class SpatialJoinModel(BaseModel):
    """Represents a spatial join model.

    Attributes
    ----------
    name : Optional[str], default None
        The name of the model.
    file : str
        The file associated with the model.
    field_name : str
        The field name used for the spatial join.
    """

    name: Optional[str] = None
    file: str
    field_name: str


class UnitSystems(str, Enum):
    """Enumeration for accepted values for the unit_system field.

    Attributes
    ----------
    imperial : str
        Represents the imperial unit system.
    metric : str
        Represents the metric unit system.
    """

    imperial = "imperial"
    metric = "metric"


class FootprintsOptions(str, Enum):
    """Enumeration for accepted values for the building_footprints field.

    Attributes
    ----------
    OSM : str
        Use OpenStreetMap for building footprints.
    """

    OSM = "OSM"


class Basins(str, Enum):
    """Enumeration class representing different basins.

    Attributes
    ----------
    NA : str
        North Atlantic
    SA : str
        South Atlantic
    EP : str
        Eastern North Pacific (which includes the Central Pacific region)
    WP : str
        Western North Pacific
    SP : str
        South Pacific
    SI : str
        South Indian
    NI : str
        North Indian
    """

    NA = "NA"
    SA = "SA"
    EP = "EP"
    WP = "WP"
    SP = "SP"
    SI = "SI"
    NI = "NI"


class GuiConfigModel(BaseModel):
    """Represents a GUI model for FloodAdapt.

    Attributes
    ----------
    max_flood_depth : float
        The last visualization bin will be ">value".
    max_aggr_dmg : float
        The last visualization bin will be ">value".
    max_footprint_dmg : float
        The last visualization bin will be ">value".
    max_benefits : float
        The last visualization bin will be ">value".
    """

    max_flood_depth: float
    max_aggr_dmg: float
    max_footprint_dmg: float
    max_benefits: float


class SviConfigModel(SpatialJoinModel):
    """Represents a model for the Social Vulnerability Index (SVI).

    Attributes
    ----------
    threshold : float
        The threshold value for the SVI model to specify vulnerability.
    """

    threshold: float


class TideGaugeConfigModel(BaseModel):
    """Represents a tide gauge model.

    Attributes
    ----------
    source : TideGaugeSource
        The source of the tide gauge data.
    description : str, default ""
        Description of the tide gauge.
    ref : Optional[str], default None
        The reference name. Should be defined in the water level references.
    id : Optional[int], default None
        The station ID.
    lon : Optional[float], default None
        Longitude of the tide gauge.
    lat : Optional[float], default None
        Latitude of the tide gauge.
    file : Optional[str], default None
        The file associated with the tide gauge data.
    max_distance : Optional[us.UnitfulLength], default None
        The maximum distance.
    """

    source: TideGaugeSource
    description: str = ""
    ref: Optional[str] = None
    id: Optional[int] = None
    lon: Optional[float] = None
    lat: Optional[float] = None
    file: Optional[str] = None
    max_distance: Optional[us.UnitfulLength] = None


class ConfigModel(BaseModel):
    """Represents the configuration model for FloodAdapt.

    Attributes
    ----------
    name : str
        The name of the site.
    description : Optional[str], default None
        The description of the site.
    database_path : Optional[str], default None
        The path to the database where all the sites are located.
    unit_system : UnitSystems
        The unit system.
    gui : GuiConfigModel
        The GUI model representing scaling values for the layers.
    infographics : Optional[bool], default True
        Indicates if infographics are enabled.
    fiat : str
        The FIAT model path.
    aggregation_areas : Optional[list[SpatialJoinModel]], default None
        The list of aggregation area models.
    building_footprints : Optional[SpatialJoinModel | FootprintsOptions], default FootprintsOptions.OSM
        The building footprints model or OSM option.
    fiat_buildings_name : str | list[str], default "buildings"
        The name(s) of the buildings geometry in the FIAT model.
    fiat_roads_name : Optional[str], default "roads"
        The name of the roads geometry in the FIAT model.
    bfe : Optional[SpatialJoinModel], default None
        The BFE model.
    svi : Optional[SviConfigModel], default None
        The SVI model.
    road_width : Optional[float], default 5
        The road width in meters.
    return_periods : list[int], default []
        The list of return periods for risk calculations.
    floodmap_type : Optional[FloodmapType], default None
        The type of floodmap to use.
    references : WaterlevelReferenceModel, default WaterlevelReferenceModel(...)
        The water level reference model.
    sfincs_overland : FloodModel
        The overland SFINCS model.
    sfincs_offshore : Optional[FloodModel], default None
        The offshore SFINCS model.
    dem : Optional[DemModel], default None
        The DEM model.
    excluded_datums : list[str], default []
        List of datums to exclude from plotting.
    slr_scenarios : Optional[SlrScenariosModel], default None
        The sea level rise scenarios model.
    scs : Optional[SCSModel], default None
        The SCS model.
    tide_gauge : Optional[TideGaugeConfigModel], default None
        The tide gauge model.
    cyclones : Optional[bool], default True
        Indicates if cyclones are enabled.
    cyclone_basin : Optional[Basins], default None
        The cyclone basin.
    obs_point : Optional[list[ObsPointModel]], default None
        The list of observation point models.
    probabilistic_set : Optional[str], default None
        The probabilistic set path.
    """

    # General
    name: str = Field(..., min_length=1, pattern='^[^<>:"/\\\\|?* ]*$')
    description: Optional[str] = None
    database_path: Optional[str] = None
    unit_system: UnitSystems
    gui: GuiConfigModel
    infographics: Optional[bool] = True

    # FIAT
    fiat: str
    aggregation_areas: Optional[list[SpatialJoinModel]] = None
    building_footprints: Optional[SpatialJoinModel | FootprintsOptions] = (
        FootprintsOptions.OSM
    )
    fiat_buildings_name: str | list[str] = "buildings"
    fiat_roads_name: Optional[str] = "roads"
    bfe: Optional[SpatialJoinModel] = None
    svi: Optional[SviConfigModel] = None
    road_width: us.UnitfulLength = us.UnitfulLength(
        value=5.0, units=us.UnitTypesLength.meters
    )
    return_periods: list[int] = Field(default_factory=list)
    floodmap_type: Optional[FloodmapType] = None

    # SFINCS
    references: Optional[WaterlevelReferenceModel] = None
    sfincs_overland: FloodModel
    sfincs_offshore: Optional[FloodModel] = None
    dem: Optional[DemModel] = None

    excluded_datums: list[str] = Field(default_factory=list)

    slr_scenarios: Optional[SlrScenariosModel] = None
    scs: Optional[SCSModel] = None
    tide_gauge: Optional[TideGaugeConfigModel] = None
    cyclones: Optional[bool] = True
    cyclone_basin: Optional[Basins] = None
    obs_point: Optional[list[ObsPointModel]] = None
    probabilistic_set: Optional[str] = None

    @staticmethod
    def read(toml_path: Union[str, Path]) -> "ConfigModel":
        """
        Read a configuration file and returns the validated attributes.

        Args:
            toml_path (str | Path): The path to the configuration file.

        Returns
        -------
            ConfigModel: The validated attributes from the configuration file.
        """
        toml_path = Path(toml_path)
        with open(toml_path, mode="rb") as fp:
            toml = tomli.load(fp)
        config = ConfigModel.model_validate(toml)

        # check if database path is provided and use config_file path if not
        if config.database_path is None:
            dbs_path = Path(toml_path).parent / "Database"
            if not dbs_path.exists():
                dbs_path.mkdir(parents=True)
            config.database_path = dbs_path.as_posix()
        # check if paths are relative to the config file and make them absolute
        config.database_path = path_check(config.database_path, toml_path)
        config.fiat = path_check(config.fiat, toml_path)
        config.sfincs_overland.name = path_check(config.sfincs_overland.name, toml_path)
        if config.sfincs_offshore:
            config.sfincs_offshore.name = path_check(
                config.sfincs_offshore.name, toml_path
            )
        if isinstance(config.building_footprints, SpatialJoinModel):
            config.building_footprints.file = path_check(
                config.building_footprints.file, toml_path
            )
        if config.tide_gauge and config.tide_gauge.file:
            config.tide_gauge.file = path_check(config.tide_gauge.file, toml_path)
        if config.svi:
            config.svi.file = path_check(config.svi.file, toml_path)
        if config.bfe:
            config.bfe.file = path_check(config.bfe.file, toml_path)
        if config.slr_scenarios:
            config.slr_scenarios.file = path_check(config.slr_scenarios.file, toml_path)
        if config.probabilistic_set:
            config.probabilistic_set = path_check(config.probabilistic_set, toml_path)
        if config.aggregation_areas:
            for aggr in config.aggregation_areas:
                aggr.file = path_check(aggr.file, toml_path)

        return config


class DatabaseBuilder:
    _has_roads: bool = False
    _aggregation_areas: Optional[list] = None
    _probabilistic_set_name: Optional[str] = None

    def __init__(self, config: ConfigModel):
        self.config = config

        # Set database root
        if config.database_path:
            self.root = Path(config.database_path).joinpath(self.config.name)
        else:
            raise ValueError(
                "Database path is not provided. Please provide a path using the 'database_path' attribute."
            )

        # Read info that needs to be used to create other models
        self.unit_system = self.create_default_units()

    @property
    def static_path(self) -> Path:
        return self.root / "static"

    @debug_timer
    def build(self, overwrite: bool = False) -> None:
        # Check if database already exists
        if self.root.exists() and not overwrite:
            raise ValueError(
                f"There is already a Database folder in '{self.root.as_posix()}'."
            )
        if self.root.exists() and overwrite:
            shutil.rmtree(self.root)
            warnings.warn(
                f"There is already a Database folder in '{self.root.as_posix()}, which will be overwritten'."
            )
        # Create database folder
        self.root.mkdir(parents=True)

        with FloodAdaptLogging.to_file(
            file_path=self.root.joinpath("database_builder.log")
        ):
            logger.info(f"Creating a FloodAdapt database in '{self.root.as_posix()}'")

            # Make folder structure and read models
            self.setup()

            # Prepare site configuration
            site = self.create_site_config()
            site.save(self.static_path / "config" / "site.toml")

            # Add infometric and infographic configurations
            self.create_infometrics()

            # Save standard objects
            self.create_standard_objects()

            # Save log file
            logger.info("FloodAdapt database creation finished!")

    @debug_timer
    def setup(self) -> None:
        # Create the models
        self.make_folder_structure()

        # Read user models and copy to templates
        self.read_template_fiat_model()
        self.read_template_sfincs_overland_model()
        self.read_template_sfincs_offshore_model()

        # Copy standard static files
        self.add_static_files()

    @debug_timer
    def set_standard_objects(self):
        # Define name and create object
        self._no_measures_strategy_name = "no_measures"
        self._current_projection_name = "current"
        if self._probabilistic_set_name is not None:
            event_list = [self._probabilistic_set_name]
        else:
            event_list = []
        std_obj = StandardObjectModel(
            events=event_list,
            projections=[self._current_projection_name],
            strategies=[self._no_measures_strategy_name],
        )
        return std_obj

    @debug_timer
    def create_standard_objects(self):
        with modified_environ(
            DATABASE_ROOT=str(self.root.parent),
            DATABASE_NAME=self.root.name,
        ):
            logger.info("Creating `no measures` strategy and `current` projection.")
            # Create database instance
            db = Database(self.root.parent, self.config.name)
            # Create no measures strategy
            strategy = Strategy(
                name=self._no_measures_strategy_name,
                measures=[],
            )
            db.strategies.save(strategy)
            # Create current projection
            projection = Projection(
                name=self._current_projection_name,
                physical_projection=PhysicalProjection(),
                socio_economic_change=SocioEconomicChange(),
            )
            db.projections.save(projection)
            # Check prob set
            if self._probabilistic_set_name is not None:
                path_toml = (
                    db.input_path
                    / "events"
                    / self._probabilistic_set_name
                    / f"{self._probabilistic_set_name}.toml"
                )
                try:
                    EventSet.load_file(path_toml)
                except Exception as e:
                    raise ValueError(
                        f"Provided probabilistic event set '{self._probabilistic_set_name}' is not valid. Error: {e}"
                    )

    ### TEMPLATE READERS ###
    @debug_timer
    def read_template_fiat_model(self):
        user_provided = self._check_exists_and_absolute(self.config.fiat)

        # Read config model
        HydromtFiatModel(root=str(user_provided), mode="r+").read()

        # Success, so copy to db and read again
        location_in_db = self.static_path / "templates" / "fiat"
        if location_in_db.exists():
            shutil.rmtree(location_in_db)
        shutil.copytree(user_provided, location_in_db)
        in_db = HydromtFiatModel(root=str(location_in_db), mode="r+")
        in_db.read()
        # Add check to make sure the geoms are correct
        # TODO this should be handled in hydromt-FIAT
        no_geoms = len(
            [name for name in in_db.config["exposure"]["geom"].keys() if "file" in name]
        )
        in_db.exposure.exposure_geoms = in_db.exposure.exposure_geoms[:no_geoms]
        in_db.exposure._geom_names = in_db.exposure._geom_names[:no_geoms]

        # Make sure that a region polygon is included
        if "region" not in in_db.geoms:
            gdf = in_db.exposure.get_full_gdf(in_db.exposure.exposure_db)
            # Combine all geometries into a single geometry
            merged_geometry = gdf.unary_union

            # If the result is not a polygon, you can create a convex hull
            if not isinstance(merged_geometry, Polygon):
                merged_geometry = merged_geometry.convex_hull
            # Create a new GeoDataFrame with the resulting polygon
            in_db.geoms["region"] = gpd.GeoDataFrame(
                geometry=[merged_geometry], crs=gdf.crs
            )

        self.fiat_model = in_db

    @debug_timer
    def read_template_sfincs_overland_model(self):
        user_provided = self._check_exists_and_absolute(
            self.config.sfincs_overland.name
        )
        user_model = HydromtSfincsModel(root=str(user_provided), mode="r")
        user_model.read()
        if user_model.crs is None:
            raise ValueError("CRS is not defined in the SFINCS model.")

        location_in_db = self.static_path / "templates" / "overland"
        if location_in_db.exists():
            shutil.rmtree(location_in_db)
        shutil.copytree(user_provided, location_in_db)
        in_db = HydromtSfincsModel(root=str(location_in_db), mode="r+")
        in_db.read()
        self.sfincs_overland_model = in_db

    @debug_timer
    def read_template_sfincs_offshore_model(self):
        if self.config.sfincs_offshore is None:
            self.sfincs_offshore_model = None
            return
        user_provided = self._check_exists_and_absolute(
            self.config.sfincs_offshore.name
        )
        user_model = HydromtSfincsModel(root=str(user_provided), mode="r+")
        user_model.read()
        if user_model.crs is None:
            raise ValueError("CRS is not defined in the SFINCS model.")
        epsg = user_model.crs.to_epsg()

        location_in_db = self.static_path / "templates" / "offshore"
        if location_in_db.exists():
            shutil.rmtree(location_in_db)
        shutil.copytree(user_provided, location_in_db)
        in_db = HydromtSfincsModel(str(location_in_db), mode="r+")
        in_db.read(epsg=epsg)
        self.sfincs_offshore_model = in_db

    ### FIAT ###
    @debug_timer
    def create_fiat_model(self) -> FiatModel:
        fiat = FiatModel(
            config=self.create_fiat_config(),
            benefits=self.create_benefit_config(),
            risk=self.create_risk_model(),
        )
        return fiat

    @debug_timer
    def create_risk_model(self) -> Optional[RiskModel]:
        # Check if return periods are provided
        if not self.config.return_periods:
            if self._probabilistic_set_name:
                risk = RiskModel()
                logger.warning(
                    f"No return periods provided, but a probabilistic set is available. Using default return periods {risk.return_periods}."
                )
                return risk
            else:
                logger.warning(
                    "No return periods provided and no probabilistic set available. Risk calculations will not be performed."
                )
                return None
        else:
            risk = RiskModel(return_periods=self.config.return_periods)
        return risk

    @debug_timer
    def create_benefit_config(self) -> Optional[BenefitsModel]:
        if self._probabilistic_set_name is None:
            logger.warning(
                "No probabilistic set found in the config, benefits will not be available."
            )
            return None
        return BenefitsModel(
            current_year=datetime.datetime.now().year,
            current_projection="current",
            baseline_strategy="no_measures",
            event_set=self._probabilistic_set_name,
        )

    @debug_timer
    def create_fiat_config(self) -> FiatConfigModel:
        # Make sure only csv objects have geometries
        self._delete_extra_geometries()

        footprints = self.create_footprints()
        if footprints is not None:
            footprints = footprints.as_posix()

        # Clip hazard and reset buildings # TODO use hydromt-FIAT instead
        if not self.fiat_model.region.empty:
            self._clip_hazard_extend()

        # Store result for possible future use in create_infographics
        self._aggregation_areas = self.create_aggregation_areas()

        roads_gpkg = self.create_roads()

        # Get classes of non-building objects
        non_buildings = ~self.fiat_model.exposure.exposure_db[
            _FIAT_COLUMNS.object_id
        ].isin(self._get_fiat_building_geoms()[_FIAT_COLUMNS.object_id])
        non_building_names = list(
            self.fiat_model.exposure.exposure_db[_FIAT_COLUMNS.primary_object_type][
                non_buildings
            ].unique()
        )

        # Update elevations
        self.update_fiat_elevation()

        self._svi = self.create_svi()

        config = FiatConfigModel(
            exposure_crs=self.fiat_model.exposure.crs,
            floodmap_type=self.read_floodmap_type(),
            bfe=self.create_bfe(),
            non_building_names=non_building_names,
            damage_unit=self.read_damage_unit(),
            building_footprints=footprints,
            roads_file_name=roads_gpkg,
            new_development_file_name=self.create_new_developments(),
            save_simulation=False,  # TODO
            infographics=self.config.infographics,
            aggregation=self._aggregation_areas,
            svi=self._svi,
        )

        # Update output geoms names
        output_geom = {}
        counter = 0
        for key in self.fiat_model.config["exposure"]["geom"].keys():
            if "file" in key:
                counter += 1
                output_geom[f"name{counter}"] = Path(
                    self.fiat_model.config["exposure"]["geom"][key]
                ).name
        self.fiat_model.config["output"]["geom"] = output_geom
        # Make sure objects are ordered based on object id
        self.fiat_model.exposure.exposure_db = (
            self.fiat_model.exposure.exposure_db.sort_values(
                by=[_FIAT_COLUMNS.object_id], ignore_index=True
            )
        )
        # Update FIAT model with the new config
        self.fiat_model.write()

        return config

    @debug_timer
    def update_fiat_elevation(self):
        """
        Update the ground elevations of FIAT objects based on the SFINCS ground elevation map.

        This method reads the DEM file and the exposure CSV file, and updates the ground elevations
        of the FIAT objects (roads and buildings) based on the nearest elevation values from the DEM.
        """
        dem_file = self._dem_path
        # TODO resolve issue with double geometries in hydromt-FIAT and use update_ground_elevation method instead
        # self.fiat_model.update_ground_elevation(dem_file, grnd_elev_unit="meters")
        logger.info(
            "Updating FIAT objects ground elevations from SFINCS ground elevation map."
        )
        # Get unit conversion factor
        SFINCS_units = us.UnitfulLength(
            value=1.0, units=us.UnitTypesLength.meters
        )  # SFINCS is always in meters
        FIAT_units = self.unit_system.default_length_units
        conversion_factor = SFINCS_units.convert(FIAT_units)

        if not math.isclose(conversion_factor, 1):
            logger.info(
                f"Ground elevation for FIAT objects is in '{FIAT_units}', while SFINCS ground elevation is in 'meters'. Values in the exposure csv will be converted by a factor of {conversion_factor}"
            )
        # Read in DEM and objects
        exposure = self.fiat_model.exposure.exposure_db
        dem = rxr.open_rasterio(dem_file)
        gdf = self._get_fiat_gdf_full()

        # Ensure gdf has the same CRS as dem
        # Determine the CRS to use for sampling
        if (
            hasattr(self.sfincs_overland_model, "crs")
            and self.sfincs_overland_model.crs is not None
        ):
            target_crs = self.sfincs_overland_model.crs
        elif (
            hasattr(dem, "rio") and hasattr(dem.rio, "crs") and dem.rio.crs is not None
        ):
            target_crs = dem.rio.crs
        else:
            target_crs = gdf.crs
            logger.warning(
                "Could not determine CRS from SFINCS model or DEM raster. Assuming the CRS is the same as the FIAT model."
            )

        if gdf.crs != target_crs:
            gdf = gdf.to_crs(target_crs)

        # Sample DEM at the centroid of each geometry
        gdf["centroid"] = gdf.geometry.centroid
        x_points = xr.DataArray(gdf["centroid"].x, dims="points")
        y_points = xr.DataArray(gdf["centroid"].y, dims="points")
        gdf["elev"] = (
            dem.sel(x=x_points, y=y_points, band=1, method="nearest").to_numpy()
            * conversion_factor
        )

        # Merge updated elevation back into exposure DataFrame
        exposure = exposure.merge(
            gdf[[_FIAT_COLUMNS.object_id, "elev"]],
            on=_FIAT_COLUMNS.object_id,
            how="left",
        )
        exposure[_FIAT_COLUMNS.ground_elevation] = exposure["elev"]
        del exposure["elev"]

        self.fiat_model.exposure.exposure_db = exposure

    def read_damage_unit(self) -> str:
        if self.fiat_model.exposure.damage_unit is None:
            logger.warning(
                "Delft-FIAT model was missing damage units so '$' was assumed."
            )
            self.fiat_model.exposure.damage_unit = "$"
        return self.fiat_model.exposure.damage_unit

    @debug_timer
    def read_floodmap_type(self) -> FloodmapType:
        if self.config.floodmap_type is not None:
            return self.config.floodmap_type
        else:
            # If there is at least on object that uses the area method, use water depths for FA calcs
            if (
                self.fiat_model.exposure.exposure_db[_FIAT_COLUMNS.extraction_method]
                == "area"
            ).any():
                return FloodmapType.water_depth
            else:
                return FloodmapType.water_level

    @debug_timer
    def create_roads(self) -> Optional[str]:
        # Make sure that FIAT roads are polygons
        if self.config.fiat_roads_name not in self.fiat_model.exposure.geom_names:
            logger.warning(
                "Road objects are not available in the FIAT model and thus would not be available in FloodAdapt."
            )
            # TODO check how this naming of output geoms should become more explicit!
            return None

        roads = self.fiat_model.exposure.exposure_geoms[self._get_fiat_road_index()]

        # TODO do we need the lanes column?
        if (
            _FIAT_COLUMNS.segment_length
            not in self.fiat_model.exposure.exposure_db.columns
        ):
            logger.warning(
                f"'{_FIAT_COLUMNS.segment_length}' column not present in the FIAT exposure csv. Road impact infometrics cannot be produced."
            )

        # TODO should this should be performed through hydromt-FIAT?
        if not isinstance(roads.geometry.iloc[0], Polygon):
            roads = roads.to_crs(roads.estimate_utm_crs())
            road_width = self.config.road_width.convert(us.UnitTypesLength.meters)
            roads.geometry = roads.geometry.buffer(road_width / 2, cap_style=2)
            roads = roads.to_crs(self.fiat_model.exposure.crs)
            self.fiat_model.exposure.exposure_geoms[self._get_fiat_road_index()] = roads
            logger.info(
                f"FIAT road objects transformed from lines to polygons assuming a road width of {self.config.road_width} meters."
            )

        self._has_roads = True
        return f"{self.config.fiat_roads_name}.gpkg"

    @debug_timer
    def create_new_developments(self) -> Optional[str]:
        return "new_development_area.gpkg"

    @debug_timer
    def create_footprints(self) -> Optional[Path]:
        if isinstance(self.config.building_footprints, SpatialJoinModel):
            # Use the provided building footprints
            building_footprints_file = self._check_exists_and_absolute(
                self.config.building_footprints.file
            )

            logger.info(
                f"Using building footprints from {Path(building_footprints_file).as_posix()}."
            )
            # Spatially join buildings and map
            # TODO use hydromt method instead
            path = self._join_building_footprints(
                self.config.building_footprints.file,
                self.config.building_footprints.field_name,
            )
            return path
        elif self.config.building_footprints == FootprintsOptions.OSM:
            logger.info(
                "Building footprint data will be downloaded from Open Street Maps."
            )
            region = self.fiat_model.region
            if region is None:
                raise ValueError(
                    "No region file found in the FIAT model. Building footprints cannot be created."
                )
            region = region.to_crs(4326)
            if isinstance(region.boundary.to_numpy()[0], MultiLineString):
                polygon = Polygon(
                    region.boundary.to_numpy()[0].envelope
                )  # TODO check if this is correct
            else:
                polygon = Polygon(region.boundary.to_numpy()[0])
            footprints = get_buildings_from_osm(polygon)
            footprints["BF_FID"] = np.arange(1, len(footprints) + 1)
            footprints = footprints[["BF_FID", "geometry"]]
            path = self._join_building_footprints(footprints, "BF_FID")
            return path
        # Then check if geometries are already footprints
        elif isinstance(
            self._get_fiat_building_geoms().geometry.iloc[0],
            (Polygon, MultiPolygon),
        ):
            logger.info(
                "Building footprints are already available in the FIAT model geometry files."
            )
            return None
        # check if it is spatially joined and/or exists already
        elif "BF_FID" in self.fiat_model.exposure.exposure_db.columns:
            add_attrs = self.fiat_model.spatial_joins["additional_attributes"]
            fiat_path = Path(self.fiat_model.root)

            if not (add_attrs and "BF_FID" in [attr["name"] for attr in add_attrs]):
                raise KeyError(
                    "While 'BF_FID' column exists, connection to a spatial footprints file is missing."
                )

            ind = [attr["name"] for attr in add_attrs].index("BF_FID")
            footprints = add_attrs[ind]
            footprints_path = fiat_path / footprints["file"]

            if not footprints_path.exists():
                raise FileNotFoundError(
                    f"While 'BF_FID' column exists, building footprints file {footprints_path} not found."
                )

            logger.info(f"Using the building footprints located at {footprints_path}.")
            return footprints_path.relative_to(self.static_path)

        # Other methods
        else:
            logger.warning(
                "No building footprints are available. Buildings will be plotted with a default shape in FloodAdapt."
            )
            return None

    @debug_timer
    def create_bfe(self) -> Optional[BFEModel]:
        if self.config.bfe is None:
            logger.warning(
                "No base flood elevation provided. Elevating building relative to base flood elevation will not be possible in FloodAdapt."
            )
            return None

        # TODO can we use hydromt-FIAT?
        bfe_file = self._check_exists_and_absolute(self.config.bfe.file)

        logger.info(
            f"Using map from {Path(bfe_file).as_posix()} as base flood elevation."
        )

        # Spatially join buildings and map
        buildings_joined, bfe = self.spatial_join(
            self._get_fiat_building_geoms(),
            bfe_file,
            self.config.bfe.field_name,
        )

        # Make sure in case of multiple values that the max is kept
        buildings_joined = (
            buildings_joined.groupby(_FIAT_COLUMNS.object_id)
            .max(self.config.bfe.field_name)
            .sort_values(by=[_FIAT_COLUMNS.object_id])
            .reset_index()
        )

        # Save the files
        fa_bfe_file = self.static_path / "bfe" / "bfe.gpkg"
        fa_bfe_file.parent.mkdir(parents=True, exist_ok=True)
        bfe.to_file(fa_bfe_file)
        csv_path = fa_bfe_file.parent / "bfe.csv"
        buildings_joined.to_csv(csv_path, index=False)

        # Save attributes
        return BFEModel(
            geom=fa_bfe_file.relative_to(self.static_path).as_posix(),
            table=csv_path.relative_to(self.static_path).as_posix(),
            field_name=self.config.bfe.field_name,
        )

    @debug_timer
    def create_aggregation_areas(self) -> list[AggregationModel]:
        # TODO split this to 3 methods?
        aggregation_areas = []

        # first check if the FIAT model has existing aggregation areas
        if self.fiat_model.spatial_joins["aggregation_areas"]:
            # Use the aggregation areas from the FIAT model
            for aggr in self.fiat_model.spatial_joins["aggregation_areas"]:
                # Check if the exposure csv has the correct column
                col_name = _FIAT_COLUMNS.aggregation_label.format(name=aggr["name"])
                if col_name not in self.fiat_model.exposure.exposure_db.columns:
                    raise KeyError(
                        f"While aggregation area '{aggr['name']}' exists in the spatial joins of the FIAT model, the column '{col_name}' is missing in the exposure csv."
                    )
                # Check equity config
                if aggr["equity"] is not None:
                    equity_config = EquityModel(
                        census_data=str(
                            self.static_path.joinpath(
                                "templates", "fiat", aggr["equity"]["census_data"]
                            )
                            .relative_to(self.static_path)
                            .as_posix()
                        ),
                        percapitaincome_label=aggr["equity"]["percapitaincome_label"],
                        totalpopulation_label=aggr["equity"]["totalpopulation_label"],
                    )
                else:
                    equity_config = None
                # Make aggregation config
                aggr = AggregationModel(
                    name=aggr["name"],
                    file=str(
                        self.static_path.joinpath("templates", "fiat", aggr["file"])
                        .relative_to(self.static_path)
                        .as_posix()
                    ),
                    field_name=aggr["field_name"],
                    equity=equity_config,
                )
                aggregation_areas.append(aggr)

                logger.info(
                    f"Aggregation areas: {aggr.name} from the FIAT model are going to be used."
                )

        # Then check if the user has provided extra aggregation areas in the config
        if self.config.aggregation_areas:
            # Loop through aggr areas given in config
            for aggr in self.config.aggregation_areas:
                # Get name of type of aggregation area
                if aggr.name is not None:
                    aggr_name = aggr.name
                else:
                    aggr_name = Path(aggr.file).stem
                # If aggregation area already in FIAT model raise Error
                if aggr_name in [aggr.name for aggr in aggregation_areas]:
                    logger.warning(
                        f"Aggregation area '{aggr_name}' already exists in the FIAT model. The input aggregation area will be ignored."
                    )
                    continue
                # Do spatial join of FIAT objects and aggregation areas
                exposure_csv = self.fiat_model.exposure.exposure_db
                gdf = self._get_fiat_gdf_full()
                gdf_joined, aggr_areas = self.spatial_join(
                    objects=gdf[[_FIAT_COLUMNS.object_id, "geometry"]],
                    layer=str(self._check_exists_and_absolute(aggr.file)),
                    field_name=aggr.field_name,
                    rename=_FIAT_COLUMNS.aggregation_label.format(name=aggr_name),
                    filter=True,
                )
                aggr_path = Path(self.fiat_model.root).joinpath(
                    "exposure", "aggregation_areas", f"{Path(aggr.file).stem}.gpkg"
                )
                aggr_path.parent.mkdir(parents=True, exist_ok=True)
                aggr_areas.to_file(aggr_path)
                exposure_csv = exposure_csv.merge(
                    gdf_joined, on=_FIAT_COLUMNS.object_id, how="left"
                )
                self.fiat_model.exposure.exposure_db = exposure_csv
                # Update spatial joins in FIAT model
                if self.fiat_model.spatial_joins["aggregation_areas"] is None:
                    self.fiat_model.spatial_joins["aggregation_areas"] = []
                self.fiat_model.spatial_joins["aggregation_areas"].append(
                    {
                        "name": aggr_name,
                        "file": aggr_path.relative_to(self.fiat_model.root).as_posix(),
                        "field_name": _FIAT_COLUMNS.aggregation_label.format(
                            name=aggr_name
                        ),
                        "equity": None,  # TODO allow adding equity as well?
                    }
                )
                # Update the aggregation areas list in the config
                aggregation_areas.append(
                    AggregationModel(
                        name=aggr_name,
                        file=aggr_path.relative_to(self.static_path).as_posix(),
                        field_name=_FIAT_COLUMNS.aggregation_label.format(
                            name=aggr_name
                        ),
                    )
                )
                logger.info(
                    f"Aggregation areas: {aggr_name} provided in the config are going to be used."
                )

        # No config provided, no aggr areas in the model -> try to use the region file as a mock aggregation area
        if (
            not self.fiat_model.spatial_joins["aggregation_areas"]
            and not self.config.aggregation_areas
        ):
            exposure_csv = self.fiat_model.exposure.exposure_db
            region = self.fiat_model.geoms["region"]
            region = region.explode().reset_index()
            region["aggr_id"] = ["region_" + str(i) for i in np.arange(len(region)) + 1]
            aggregation_path = Path(self.fiat_model.root).joinpath(
                "aggregation_areas", "region.geojson"
            )
            if not aggregation_path.parent.exists():
                aggregation_path.parent.mkdir()

            region.to_file(aggregation_path)
            aggr = AggregationModel(
                name="region",
                file=str(aggregation_path.relative_to(self.static_path).as_posix()),
                field_name="aggr_id",
            )
            aggregation_areas.append(aggr)

            # Add column in FIAT
            gdf = self._get_fiat_gdf_full()
            gdf_joined, aggr_areas = self.spatial_join(
                objects=gdf[[_FIAT_COLUMNS.object_id, "geometry"]],
                layer=region,
                field_name="aggr_id",
                rename=_FIAT_COLUMNS.aggregation_label.format(name="region"),
            )
            exposure_csv = exposure_csv.merge(
                gdf_joined, on=_FIAT_COLUMNS.object_id, how="left"
            )
            self.fiat_model.exposure.exposure_db = exposure_csv
            logger.warning(
                "No aggregation areas were available in the FIAT model and none were provided in the config file. The region file will be used as a mock aggregation area."
            )
        return aggregation_areas

    @debug_timer
    def create_svi(self) -> Optional[SVIModel]:
        if self.config.svi:
            svi_file = self._check_exists_and_absolute(self.config.svi.file)
            exposure_csv = self.fiat_model.exposure.exposure_db
            buildings_joined, svi = self.spatial_join(
                self._get_fiat_building_geoms(),
                svi_file,
                self.config.svi.field_name,
                rename="SVI",
                filter=True,
            )
            # Add column to exposure
            if "SVI" in exposure_csv.columns:
                logger.info(
                    f"'SVI' column in the FIAT exposure csv will be replaced by {svi_file.as_posix()}."
                )
                del exposure_csv["SVI"]
            else:
                logger.info(
                    f"'SVI' column in the FIAT exposure csv will be filled by {svi_file.as_posix()}."
                )
            exposure_csv = exposure_csv.merge(
                buildings_joined, on=_FIAT_COLUMNS.object_id, how="left"
            )
            self.fiat_model.exposure.exposure_db = exposure_csv

            # Save the spatial file for future use
            svi_path = self.static_path / "templates" / "fiat" / "svi" / "svi.gpkg"
            svi_path.parent.mkdir(parents=True, exist_ok=True)
            svi.to_file(svi_path)
            logger.info(
                f"An SVI map can be shown in FloodAdapt GUI using '{self.config.svi.field_name}' column from {svi_file.as_posix()}"
            )

            return SVIModel(
                geom=Path(svi_path.relative_to(self.static_path)).as_posix(),
                field_name="SVI",
            )
        elif "SVI" in self.fiat_model.exposure.exposure_db.columns:
            logger.info(
                "'SVI' column present in the FIAT exposure csv. Vulnerability type infometrics can be produced."
            )
            add_attrs = self.fiat_model.spatial_joins["additional_attributes"]
            if "SVI" not in [attr["name"] for attr in add_attrs]:
                logger.warning("No SVI map found to display in the FloodAdapt GUI!")

            ind = [attr["name"] for attr in add_attrs].index("SVI")
            svi = add_attrs[ind]
            svi_path = self.static_path / "templates" / "fiat" / svi["file"]
            logger.info(
                f"An SVI map can be shown in FloodAdapt GUI using '{svi['field_name']}' column from {svi['file']}"
            )
            # Save site attributes
            return SVIModel(
                geom=Path(svi_path.relative_to(self.static_path)).as_posix(),
                field_name=svi["field_name"],
            )

        else:
            logger.warning(
                "'SVI' column not present in the FIAT exposure csv. Vulnerability type infometrics cannot be produced."
            )
            return None

    ### SFINCS ###
    @debug_timer
    def create_sfincs_config(self) -> SfincsModel:
        # call these functions before others to make sure water level references are updated
        config = self.create_sfincs_model_config()
        self.water_level_references = self.create_water_level_references()
        self.tide_gauge = self.create_tide_gauge()

        sfincs = SfincsModel(
            config=config,
            water_level=self.water_level_references,
            slr_scenarios=self.create_slr(),
            dem=self.create_dem_model(),
            scs=self.create_scs_model(),
            cyclone_track_database=self.create_cyclone_track_database(),
            tide_gauge=self.tide_gauge,
            river=self.create_rivers(),
            obs_point=self.create_observation_points(),
        )

        return sfincs

    @debug_timer
    def create_water_level_references(self) -> WaterlevelReferenceModel:
        sfincs_ref = self.config.sfincs_overland.reference
        if self.config.references is None:
            logger.warning(
                f"No water level references provided in the config file. Using reference provided for overland SFINCS model '{sfincs_ref}' as the main reference."
            )
            refs = WaterlevelReferenceModel(
                reference=sfincs_ref,
                datums=[
                    DatumModel(
                        name=sfincs_ref,
                        height=us.UnitfulLength(
                            value=0.0, units=self.unit_system.default_length_units
                        ),
                    )
                ],
            )
        else:
            # Check if sfincs_ref is in the references
            if sfincs_ref not in [ref.name for ref in self.config.references.datums]:
                raise ValueError(
                    f"Reference '{sfincs_ref}' not found in the provided references."
                )
            else:
                refs = self.config.references

        return refs

    @debug_timer
    def create_cyclone_track_database(self) -> Optional[CycloneTrackDatabaseModel]:
        if not self.config.cyclones or not self.config.sfincs_offshore:
            logger.warning("No cyclones will be available in the database.")
            return None

        if self.config.cyclone_basin:
            basin = self.config.cyclone_basin
        else:
            basin = "ALL"

        name = f"IBTrACS.{basin.value}.v04r01.nc"
        url = f"https://www.ncei.noaa.gov/data/international-best-track-archive-for-climate-stewardship-ibtracs/v04r01/access/netcdf/{name}"
        logger.info(f"Downloading cyclone track database from {url}")
        fn = Path(self.root) / "static" / "cyclone_track_database" / name
        fn.parent.mkdir(parents=True, exist_ok=True)

        try:
            urlretrieve(url, fn)
        except Exception:
            logger.warning(f"Could not retrieve cyclone track database from {url}")
            logger.warning("No cyclones will be available in the database.")
            return None

        return CycloneTrackDatabaseModel(file=name)

    @debug_timer
    def create_scs_model(self) -> Optional[SCSModel]:
        if self.config.scs is None:
            return None
        scs_file = self._check_exists_and_absolute(self.config.scs.file)
        db_scs_file = self.static_path / "scs" / scs_file.name
        db_scs_file.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy2(scs_file, db_scs_file)

        return SCSModel(file=scs_file.name, type=self.config.scs.type)

    @debug_timer
    def create_dem_model(self) -> DemModel:
        if self.config.dem:
            subgrid_sfincs = Path(self.config.dem.filename)
            delete_sfincs_folder = False
        else:
            logger.warning(
                "No subgrid depth geotiff file provided in the config file. Using the one from the SFINCS model."
            )
            subgrid_sfincs_folder = Path(self.sfincs_overland_model.root) / "subgrid"
            subgrid_sfincs = subgrid_sfincs_folder / "dep_subgrid.tif"
            delete_sfincs_folder = True

        dem_file = self._check_exists_and_absolute(subgrid_sfincs)
        fa_subgrid_path = self.static_path / "dem" / dem_file.name
        fa_subgrid_path.parent.mkdir(parents=True, exist_ok=True)

        # Check tiles
        tiles_sfincs = Path(self.sfincs_overland_model.root) / "tiles"
        fa_tiles_path = self.static_path / "dem" / "tiles"
        if tiles_sfincs.exists():
            shutil.move(tiles_sfincs, fa_tiles_path)
            if (fa_tiles_path / "index").exists():
                os.rename(fa_tiles_path / "index", fa_tiles_path / "indices")
            logger.info(
                "Tiles were already available in the SFINCS model and will directly be used in FloodAdapt."
            )
        else:
            # Make tiles
            fa_tiles_path.mkdir(parents=True)
            self.sfincs_overland_model.setup_tiles(
                path=fa_tiles_path,
                datasets_dep=[{"elevtn": dem_file}],
                zoom_range=[0, 13],
                fmt="png",
            )
            logger.info(
                f"Tiles were created using the {subgrid_sfincs.as_posix()} as the elevation map."
            )

        shutil.copy2(dem_file, fa_subgrid_path)
        self._dem_path = fa_subgrid_path

        # Remove the original subgrid folder if it exists
        if delete_sfincs_folder:
            gc.collect()
            if subgrid_sfincs_folder.exists() and subgrid_sfincs_folder.is_dir():
                shutil.rmtree(subgrid_sfincs_folder)

        return DemModel(
            filename=fa_subgrid_path.name, units=us.UnitTypesLength.meters
        )  # always in meters

    @debug_timer
    def create_sfincs_model_config(self) -> SfincsConfigModel:
        config = SfincsConfigModel(
            csname=self.sfincs_overland_model.crs.name,
            cstype=Cstype(
                self.sfincs_overland_model.crs.type_name.split(" ")[0].lower()
            ),
            offshore_model=self.create_offshore_model(),
            overland_model=self.create_overland_model(),
            floodmap_units=self.unit_system.default_length_units,
            save_simulation=False,
        )

        return config

    @debug_timer
    def create_slr(self) -> Optional[SlrScenariosModel]:
        if self.config.slr_scenarios is None:
            return None

        self.config.slr_scenarios.file = str(
            self._check_exists_and_absolute(self.config.slr_scenarios.file)
        )
        slr_path = self.static_path / "slr_scenarios"
        slr_path.mkdir()
        new_file = slr_path / Path(self.config.slr_scenarios.file).name
        shutil.copyfile(self.config.slr_scenarios.file, new_file)

        return SlrScenariosModel(
            file=new_file.relative_to(self.static_path).as_posix(),
            relative_to_year=self.config.slr_scenarios.relative_to_year,
        )

    @debug_timer
    def create_observation_points(self) -> Union[list[ObsPointModel], None]:
        if self.config.obs_point is None:
            obs_points = []
        else:
            logger.info("Observation points were provided in the config file.")
            obs_points = self.config.obs_point
        if self.tide_gauge is not None:
            # Check if the tide gauge point is within the SFINCS region
            region = self.sfincs_overland_model.region
            point = gpd.GeoSeries(
                [gpd.points_from_xy([self.tide_gauge.lon], [self.tide_gauge.lat])[0]],
                crs=4326,
            )
            region_4326 = region.to_crs(4326)
            if not point.within(region_4326.unary_union).item():
                logger.warning(
                    "The tide gauge location is outside the SFINCS region and will not be added as an observation point."
                )
            else:
                logger.info(
                    "A tide gauge has been setup in the database. It will be used as an observation point as well."
                )
                obs_points.append(
                    ObsPointModel(
                        name=self.tide_gauge.name,
                        description="Tide gauge observation point",
                        ID=self.tide_gauge.ID,
                        lon=self.tide_gauge.lon,
                        lat=self.tide_gauge.lat,
                    )
                )

        if not obs_points:
            logger.warning(
                "No observation points were provided in the config file or created from the tide gauge. No observation points will be available in FloodAdapt."
            )
            return None
        else:
            return obs_points

    @debug_timer
    def create_rivers(self) -> list[RiverModel]:
        src_file = Path(self.sfincs_overland_model.root) / "sfincs.src"
        if not src_file.exists():
            logger.warning("No rivers found in the SFINCS model.")
            return []

        df = pd.read_csv(src_file, delim_whitespace=True, header=None, names=["x", "y"])
        river_locs = gpd.GeoDataFrame(
            df,
            geometry=gpd.points_from_xy(df.x, df.y),
            crs=self.sfincs_overland_model.crs,
        )
        rivers = []
        for idx, row in river_locs.iterrows():
            if "dis" in self.sfincs_overland_model.forcing:
                discharge = (
                    self.sfincs_overland_model.forcing["dis"]
                    .sel(index=idx + 1)
                    .to_numpy()
                    .mean()
                )
            else:
                discharge = 0
                logger.warning(
                    f"No river discharge conditions were found in the SFINCS model for river {idx}. A default value of 0 will be used."
                )

            river = RiverModel(
                name=f"river_{idx}",
                x_coordinate=row.x,
                y_coordinate=row.y,
                mean_discharge=us.UnitfulDischarge(
                    value=discharge, units=self.unit_system.default_discharge_units
                ),
            )
            rivers.append(river)

        logger.info(
            f"{len(river_locs)} river(s) were identified from the SFINCS model and will be available in FloodAdapt for discharge input."
        )

        return rivers

    @debug_timer
    def create_tide_gauge(self) -> Optional[TideGauge]:
        if self.config.tide_gauge is None:
            logger.warning(
                "Tide gauge information not provided. Historical events will not have an option to use gauged data in FloodAdapt!"
            )
            return None

        if self.config.tide_gauge.source == TideGaugeSource.file:
            if self.config.tide_gauge.file is None:
                raise ValueError(
                    "Tide gauge file needs to be provided when 'file' is selected as the source."
                )
            if self.config.tide_gauge.ref is None:
                logger.warning(
                    f"Tide gauge reference not provided. '{self.water_level_references.reference}' is assumed as the reference of the water levels in the file."
                )
                self.config.tide_gauge.ref = self.water_level_references.reference
            else:
                if self.config.tide_gauge.ref not in [
                    datum.name for datum in self.water_level_references.datums
                ]:
                    raise ValueError(
                        f"Provided tide gauge reference '{self.config.tide_gauge.ref}' not found in the water level references!"
                    )

            tide_gauge_file = self._check_exists_and_absolute(
                self.config.tide_gauge.file
            )
            db_file_path = Path(self.static_path / "tide_gauges") / tide_gauge_file.name

            db_file_path.parent.mkdir(parents=True, exist_ok=True)
            shutil.copyfile(self.config.tide_gauge.file, db_file_path)

            rel_db_path = Path(db_file_path.relative_to(self.static_path))
            logger.warning(
                f"Tide gauge from file {rel_db_path} assumed to be in {self.unit_system.default_length_units}!"
            )
            tide_gauge = TideGauge(
                reference=self.config.tide_gauge.ref,
                description="Observations from file stored in database",
                source=TideGaugeSource.file,
                file=rel_db_path,
                lon=self.config.tide_gauge.lon,
                lat=self.config.tide_gauge.lat,
                units=self.unit_system.default_length_units,
            )

            return tide_gauge

        elif self.config.tide_gauge.source == TideGaugeSource.noaa_coops:
            if self.config.tide_gauge.ref is not None:
                ref = self.config.tide_gauge.ref
            else:
                ref = "MLLW"  # If reference is not provided use MLLW

            if self.config.tide_gauge.id is None:
                station_id = self._get_closest_station()
                logger.info(
                    "The closest NOAA tide gauge station to the site will be searched."
                )
            else:
                station_id = self.config.tide_gauge.id
                logger.info(
                    f"The NOAA tide gauge station with the provided ID {station_id} will be used."
                )
            station = self._get_station_metadata(station_id=station_id, ref=ref)
            if station is not None:
                # First create water level references based on station
                # Get datums
                datums = []
                # Get local datum
                datums.append(
                    DatumModel(
                        name=station["datum_name"],
                        height=us.UnitfulLength(
                            value=station["datum"], units=station["units"]
                        ).transform(self.unit_system.default_length_units),
                    )
                )
                # Get MSL
                datums.append(
                    DatumModel(
                        name="MSL",
                        height=us.UnitfulLength(
                            value=station["msl"], units=station["units"]
                        ).transform(self.unit_system.default_length_units),
                    )
                )
                # Get extras
                for name in ["MLLW", "MHHW"]:
                    height = us.UnitfulLength(
                        value=station[name.lower()], units=station["units"]
                    ).transform(self.unit_system.default_length_units)

                    wl_info = DatumModel(
                        name=name,
                        height=height,
                    )
                    datums.append(wl_info)

                station_refs = WaterlevelReferenceModel(reference=ref, datums=datums)

                # Check if we can translate the rest of the datums
                if self.water_level_references.reference != station_refs.reference:
                    for dat in self.water_level_references.datums:
                        if dat.name not in [
                            datum.name for datum in station_refs.datums
                        ]:
                            # If datum is not in the datums, try to convert it
                            h1 = dat.height
                            ref1 = self.water_level_references.reference
                            h2 = h1 + station_refs.get_datum(ref1).height
                            # Replace the datum in self.water_level_references.datums
                            dat.height = h2
                            logger.warning(
                                f"Datum '{dat.name}' converted to reference '{ref1}' with new height {h2}."
                            )

                # Check if datums already exist in the water level references and replace
                for datum in datums:
                    existing_datum = next(
                        (
                            dat
                            for dat in self.water_level_references.datums
                            if dat.name == datum.name
                        ),
                        None,
                    )
                    if existing_datum:
                        self.water_level_references.datums.remove(existing_datum)
                        logger.warning(
                            f"Datum '{datum.name}' already exists in config reference. Replacing it based on NOAA station data."
                        )
                    self.water_level_references.datums.append(datum)

                # Update reference datum
                self.water_level_references.reference = (
                    ref  # update the water level reference
                )
                logger.warning(f"Main water level reference set to '{ref}'.")

                # Add tide_gauge information in site toml
                tide_gauge = TideGauge(
                    name=station["name"],
                    description=f"observations from '{self.config.tide_gauge.source}' api",
                    source=self.config.tide_gauge.source,
                    reference=ref,
                    ID=int(station["id"]),
                    lon=station["lon"],
                    lat=station["lat"],
                    units=us.UnitTypesLength.meters,  # the api always asks for SI units right now
                )

            return tide_gauge
        else:
            logger.warning(
                f"Tide gauge source not recognized: {self.config.tide_gauge.source}. Historical events will not have an option to use gauged data in FloodAdapt!"
            )
            return None

    @debug_timer
    def create_offshore_model(self) -> Optional[FloodModel]:
        if self.sfincs_offshore_model is None:
            return None
        # Connect boundary points of overland to output points of offshore
        fn = Path(self.sfincs_overland_model.root) / "sfincs.bnd"
        bnd = pd.read_csv(fn, sep=" ", lineterminator="\n", header=None)
        bnd = bnd.rename(columns={0: "x", 1: "y"})
        bnd_geo = gpd.GeoDataFrame(
            bnd,
            geometry=gpd.points_from_xy(bnd.x, bnd.y),
            crs=self.sfincs_overland_model.config["epsg"],
        )
        obs_geo = bnd_geo.to_crs(4326)
        obs_geo["x"] = obs_geo.geometry.x
        obs_geo["y"] = obs_geo.geometry.y
        del obs_geo["geometry"]
        obs_geo["name"] = [f"bnd_pt{num:02d}" for num in range(1, len(obs_geo) + 1)]
        fn_off = Path(self.sfincs_offshore_model.root) / "sfincs.obs"
        obs_geo.to_csv(
            fn_off,
            sep="\t",
            index=False,
            header=False,
        )
        logger.info(
            "Output points of the offshore SFINCS model were reconfigured to the boundary points of the overland SFINCS model."
        )

        return FloodModel(
            name="offshore",
            reference=self.config.sfincs_offshore.reference,
            vertical_offset=self.config.sfincs_offshore.vertical_offset,
        )

    @debug_timer
    def create_overland_model(self) -> FloodModel:
        return FloodModel(
            name="overland",
            reference=self.config.sfincs_overland.reference,
        )

    ### SITE ###
    @debug_timer
    def create_site_config(self) -> Site:
        """Create the site configuration for the FloodAdapt model.

        The order of these functions is important!
        1. Create the SFINCS model.
            needs: water level references
            provides: updated water level references with optional tide gauge
        2. Create the FIAT model.
            needs: water level references and optional probabilistic event set
            provides: svi and exposure geometries
        3. Create the GUI model. (requires water level references and FIAT model to be updated)
            needs: water level references and FIAT model to be updated
            provides: gui model with output layers, visualization layers and plotting.

        """
        sfincs = self.create_sfincs_config()
        self.add_probabilistic_set()
        fiat = self.create_fiat_model()
        gui = self.create_gui_config()

        # Order doesnt matter from here
        lon, lat = self.read_location()
        std_objs = self.set_standard_objects()
        description = (
            self.config.description if self.config.description else self.config.name
        )

        config = Site(
            name=self.config.name,
            description=description,
            lat=lat,
            lon=lon,
            fiat=fiat,
            gui=gui,
            sfincs=sfincs,
            standard_objects=std_objs,
        )
        return config

    @debug_timer
    def read_location(self) -> tuple[float, float]:
        # Get center of area of interest
        if not self.fiat_model.region.empty:
            center = self.fiat_model.region.dissolve().centroid.to_crs(4326)[0]
        else:
            center = self._get_fiat_building_geoms().dissolve().centroid.to_crs(4326)[0]
        return center.x, center.y

    @debug_timer
    def create_gui_config(self) -> GuiModel:
        gui = GuiModel(
            units=self.unit_system,
            plotting=self.create_hazard_plotting_config(),
            output_layers=self.create_output_layers_config(),
            visualization_layers=self.create_visualization_layers(),
        )

        return gui

    @debug_timer
    def create_default_units(self) -> GuiUnitModel:
        if self.config.unit_system == UnitSystems.imperial:
            return GuiUnitModel.imperial()
        elif self.config.unit_system == UnitSystems.metric:
            return GuiUnitModel.metric()
        else:
            raise ValueError(
                f"Unit system {self.config.unit_system} not recognized. Please choose 'imperial' or 'metric'."
            )

    @debug_timer
    def create_visualization_layers(self) -> VisualizationLayers:
        visualization_layers = VisualizationLayers()
        if self._svi is not None:
            visualization_layers.add_layer(
                name="svi",
                long_name="Social Vulnerability Index (SVI)",
                path=str(self.static_path / self._svi.geom),
                database_path=self.root,
                field_name="SVI",
                bins=[0.05, 0.2, 0.4, 0.6, 0.8],
            )
        return visualization_layers

    @debug_timer
    def create_output_layers_config(self) -> OutputLayers:
        # Read default colors from template
        fd_max = self.config.gui.max_flood_depth
        ad_max = self.config.gui.max_aggr_dmg
        ftd_max = self.config.gui.max_footprint_dmg
        b_max = self.config.gui.max_benefits

        benefits_layer = None
        if self.config.probabilistic_set is not None:
            benefits_layer = BenefitsLayer(
                bins=[0, 0.01, 0.02 * b_max, 0.2 * b_max, b_max],
                colors=[
                    "#FF7D7D",
                    "#FFFFFF",
                    "#DCEDC8",
                    "#AED581",
                    "#7CB342",
                    "#33691E",
                ],
                threshold=0.0,
            )

        output_layers = OutputLayers(
            floodmap=FloodMapLayer(
                bins=[0.2 * fd_max, 0.6 * fd_max, fd_max],
                colors=["#D7ECFB", "#8ABDDD", "#1C73A4", "#081D58"],
                zbmax=-9999,
                depth_min=0.0,
            ),
            aggregation_dmg=AggregationDmgLayer(
                bins=[0.00001, 0.1 * ad_max, 0.25 * ad_max, 0.5 * ad_max, ad_max],
                colors=[
                    "#FFFFFF",
                    "#FEE9CE",
                    "#FDBB84",
                    "#FC844E",
                    "#E03720",
                    "#860000",
                ],
            ),
            footprints_dmg=FootprintsDmgLayer(
                bins=[0.00001, 0.06 * ftd_max, 0.2 * ftd_max, 0.4 * ftd_max, ftd_max],
                colors=[
                    "#FFFFFF",
                    "#FEE9CE",
                    "#FDBB84",
                    "#FC844E",
                    "#E03720",
                    "#860000",
                ],
            ),
            benefits=benefits_layer,
        )
        return output_layers

    @debug_timer
    def create_hazard_plotting_config(self) -> PlottingModel:
        datum_names = [datum.name for datum in self.water_level_references.datums]
        if "MHHW" in datum_names:
            amplitude = (
                self.water_level_references.get_datum("MHHW").height
                - self.water_level_references.get_datum("MSL").height
            )
            logger.info(
                f"The default tidal amplitude in the GUI will be {amplitude.transform(self.unit_system.default_length_units)}, calculated as the difference between MHHW and MSL from the tide gauge data."
            )
        else:
            amplitude = us.UnitfulLength(
                value=0.0, units=self.unit_system.default_length_units
            )
            logger.warning(
                "The default tidal amplitude in the GUI will be 0.0, since no tide-gauge water levels are available. You can change this in the site.toml with the 'gui.tide_harmonic_amplitude' attribute."
            )

        ref = "MSL"
        if ref not in datum_names:
            logger.warning(
                f"The Mean Sea Level (MSL) datum is not available in the site.toml. The synthetic tide will be created relative to the main reference: {self.water_level_references.reference}."
            )
            ref = self.water_level_references.reference

        plotting = PlottingModel(
            synthetic_tide=SyntheticTideModel(
                harmonic_amplitude=amplitude,
                datum=ref,
            ),
            excluded_datums=self.config.excluded_datums,
        )

        return plotting

    @debug_timer
    def create_infometrics(self):
        """
        Copy the infometrics and infographics templates to the appropriate location and modifies the metrics_config.toml files.

        This method copies the templates from the 'infometrics' and 'infographics' folders to the 'static/templates' folder in the root directory.
        It then modifies the 'metrics_config.toml' and 'metrics_config_risk.toml' files by updating the 'aggregateBy' attribute with the names
        of the aggregations defined in the 'fiat' section of the 'site_attrs' attribute.
        """
        # TODO there should be generalized infometric queries with NSI or OSM, and with SVI or without. Then Based on the user input these should be chosen automatically
        templates_path = Path(__file__).parent.resolve().joinpath("templates")

        # Create template folder
        path_im = self.root.joinpath("static", "templates", "infometrics")
        path_im.mkdir()

        # Copy mandatory metric configs
        path_im_temp = templates_path.joinpath("infometrics")
        for file in path_im_temp.glob("*.toml"):
            shutil.copy2(file, path_im)

        self._create_optional_infometrics(templates_path, path_im)

        files = list(path_im.glob("*metrics_config*.toml"))
        # Update aggregation areas in metrics config
        for file in files:
            file = path_im.joinpath(file)
            with open(file, "rb") as f:
                attrs = tomli.load(f)

            # add aggration levels
            if self._aggregation_areas is None:
                self._aggregation_areas = self.create_aggregation_areas()
            attrs["aggregateBy"] = [aggr.name for aggr in self._aggregation_areas]

            # take out road metrics if needed
            if not self._has_roads:
                attrs["queries"] = [
                    query
                    for query in attrs["queries"]
                    if "road" not in query["name"].lower()
                ]

            # Replace Damage Unit
            # TODO do this in a better manner
            for i, query in enumerate(attrs["queries"]):
                if "$" in query["long_name"]:
                    query["long_name"] = query["long_name"].replace(
                        "$", self.read_damage_unit()
                    )

            # replace the SVI threshold if needed
            if self.config.svi:
                for i, query in enumerate(attrs["queries"]):
                    query["filter"] = query["filter"].replace(
                        "SVI_threshold", str(self.config.svi.threshold)
                    )

            with open(file, "wb") as f:
                tomli_w.dump(attrs, f)

    @debug_timer
    def _create_optional_infometrics(self, templates_path: Path, path_im: Path):
        # If infographics are going to be created in FA, get template metric configurations
        if not self.config.infographics:
            return

        # Check what type of infographics should be used
        if self.config.unit_system == UnitSystems.imperial:
            metrics_folder_name = "US_NSI"
            logger.info("Default NSI infometrics and infographics will be created.")
        elif self.config.unit_system == UnitSystems.metric:
            metrics_folder_name = "OSM"
            logger.info("Default OSM infometrics and infographics will be created.")
        else:
            raise ValueError(
                f"Unit system {self.config.unit_system} is not recognized. Please choose 'imperial' or 'metric'."
            )

        if self.config.svi is not None:
            svi_folder_name = "with_SVI"
        else:
            svi_folder_name = "without_SVI"

        # Copy metrics config for infographics
        path_0 = templates_path.joinpath(
            "infometrics", metrics_folder_name, svi_folder_name
        )
        for file in path_0.glob("*.toml"):
            shutil.copy2(file, path_im)

        # Copy additional risk config
        file = templates_path.joinpath(
            "infometrics",
            metrics_folder_name,
            "metrics_additional_risk_configs.toml",
        )
        shutil.copy2(file, path_im)

        # Copy infographics config
        path_ig_temp = templates_path.joinpath("infographics", metrics_folder_name)
        path_ig = self.root.joinpath("static", "templates", "infographics")
        path_ig.mkdir()
        files_ig = ["styles.css", "config_charts.toml"]

        if self.config.svi is not None:
            files_ig.append("config_risk_charts.toml")
            files_ig.append("config_people.toml")

        if self._has_roads:
            files_ig.append("config_roads.toml")

        for file in files_ig:
            shutil.copy2(path_ig_temp.joinpath(file), path_ig.joinpath(file))

        # Copy images
        path_0 = templates_path.joinpath("infographics", "images")
        path_1 = self.root.joinpath("static", "templates", "infographics", "images")
        shutil.copytree(path_0, path_1)

    @debug_timer
    def add_static_files(self):
        """
        Copy static files from the 'templates' folder to the 'static' folder.

        This method iterates over a list of folders and copies the contents of each folder from the 'templates' directory
        to the corresponding folder in the 'static' directory.
        """
        templates_path = Path(__file__).parent.resolve().joinpath("templates")
        folders = ["icons", "green_infra_table"]
        for folder in folders:
            path_0 = templates_path.joinpath(folder)
            path_1 = self.static_path / folder
            shutil.copytree(path_0, path_1)

        # Check table values
        green_infra_path = (
            self.static_path / "green_infra_table" / "green_infra_lookup_table.csv"
        )
        df = pd.read_csv(green_infra_path)
        # Convert "Infiltration depth (feet)" to the database unit system and rename column
        # Find the column that has "Infiltration depth" in its name
        infiltration_col = next(
            (col for col in df.columns if "Infiltration depth" in col), None
        )
        # Try to infer the units from the column name, e.g., "Infiltration depth (feet)"
        match = re.search(r"\((.*?)\)", infiltration_col)
        current_units = match.group(1).strip()

        # Determine target units and column name
        if self.unit_system.default_length_units != current_units:
            target_units = self.unit_system.default_length_units
            new_col = f"Infiltration depth ({target_units.value})"
            conversion_factor = us.UnitfulLength(
                value=1.0, units=current_units
            ).convert(target_units)

            df[new_col] = (df[infiltration_col] * conversion_factor).round(2)
            df = df.drop(columns=[infiltration_col])
            # Save the updated table
            df.to_csv(green_infra_path, index=False)

    @debug_timer
    def add_probabilistic_set(self):
        # Copy prob set if given
        if self.config.probabilistic_set:
            logger.info(
                f"Probabilistic event set imported from {self.config.probabilistic_set}"
            )
            prob_event_name = Path(self.config.probabilistic_set).name
            path_db = self.root.joinpath("input", "events", prob_event_name)
            shutil.copytree(self.config.probabilistic_set, path_db)
            self._probabilistic_set_name = prob_event_name
        else:
            logger.warning(
                "Probabilistic event set not provided. Risk scenarios cannot be run in FloodAdapt."
            )
            self._probabilistic_set_name = None

    ### HELPER FUNCTIONS ###
    def make_folder_structure(self) -> None:
        """
        Create the folder structure for the database.

        This method creates the necessary folder structure for the FloodAdapt database, including
        the input and static folders. It also creates subfolders within the input and
        static folders based on a predefined list of names.
        """
        logger.info("Preparing the database folder structure.")
        inputs = [
            "events",
            "projections",
            "measures",
            "strategies",
            "scenarios",
            "benefits",
        ]
        for name in inputs:
            (self.root / "input" / name).mkdir(parents=True, exist_ok=True)

        # Prepare static folder structure
        folders = ["templates", "config"]
        for name in folders:
            (self.static_path / name).mkdir(parents=True, exist_ok=True)

    def _check_exists_and_absolute(self, path: str) -> Path:
        """Check if the path is absolute or relative and return a Path object. Raises an error if the path is not valid."""
        if not Path(path).exists():
            raise FileNotFoundError(f"Path {path} does not exist.")

        if Path(path).is_absolute():
            return Path(path)
        else:
            raise ValueError(f"Path {path} is not absolute.")

    def _get_fiat_building_geoms(self) -> gpd.GeoDataFrame:
        """
        Get the building geometries from the FIAT model.

        Returns
        -------
        gpd.GeoDataFrame
            A GeoDataFrame containing the building geometries.
        """
        building_indices = self._get_fiat_building_index()
        buildings = pd.concat(
            [self.fiat_model.exposure.exposure_geoms[i] for i in building_indices],
            ignore_index=True,
        )
        return buildings

    @debug_timer
    def _join_building_footprints(
        self, building_footprints: gpd.GeoDataFrame, field_name: str
    ) -> Path:
        """
        Join building footprints with existing building data and updates the exposure CSV.

        Args:
            building_footprints (GeoDataFrame): GeoDataFrame containing the building footprints to be joined.
            field_name (str): The field name to use for the spatial join.

        Returns
        -------
        This method performs the following steps:
        1. Reads the exposure CSV file.
        2. Performs a spatial join between the buildings and building footprints.
        3. Ensures that in case of multiple values, the first is kept.
        4. Creates a folder to store the building footprints.
        5. Saves the spatial file for future use.
        6. Merges the joined buildings with the exposure CSV and saves it.
        7. Updates the site attributes with the relative path to the saved building footprints.
        8. Logs the location where the building footprints are saved.
        """
        buildings = self._get_fiat_building_geoms()
        exposure_csv = self.fiat_model.exposure.exposure_db
        if "BF_FID" in exposure_csv.columns:
            logger.warning(
                "Column 'BF_FID' already exists in the exposure columns and will be replaced."
            )
            del exposure_csv["BF_FID"]
        buildings_joined, building_footprints = self.spatial_join(
            buildings,
            building_footprints,
            field_name,
            rename="BF_FID",
            filter=True,
        )
        # Make sure in case of multiple values that the first is kept
        buildings_joined = (
            buildings_joined.groupby(_FIAT_COLUMNS.object_id)
            .first()
            .sort_values(by=[_FIAT_COLUMNS.object_id])
        )
        # Create folder
        bf_folder = Path(self.fiat_model.root) / "exposure" / "building_footprints"
        bf_folder.mkdir(parents=True, exist_ok=True)

        # Save the spatial file for future use
        geo_path = bf_folder / "building_footprints.gpkg"
        building_footprints.to_file(geo_path)

        # Save to exposure csv
        exposure_csv = exposure_csv.merge(
            buildings_joined, on=_FIAT_COLUMNS.object_id, how="left"
        )

        # Set model building footprints
        self.fiat_model.building_footprint = building_footprints
        self.fiat_model.exposure.exposure_db = exposure_csv

        # Save site attributes
        buildings_path = geo_path.relative_to(self.static_path)
        logger.info(
            f"Building footprints saved at {(self.static_path / buildings_path).resolve().as_posix()}"
        )

        return buildings_path

    @debug_timer
    def _clip_hazard_extend(self, clip_footprints=True):
        """
        Clip the exposure data to the bounding box of the hazard data.

        This method clips the exposure data to the bounding box of the hazard data. It creates a GeoDataFrame
        from the hazard polygons, and then uses the `gpd.clip` function to clip the exposure geometries to the
        bounding box of the hazard polygons. If the exposure data contains roads, it is split into two separate
        GeoDataFrames: one for buildings and one for roads. The clipped exposure data is then saved back to the
        `exposure_db` attribute of the `FiatModel` object.

        Parameters
        ----------
        clip_footprints : bool, default True
            Whether to clip the building footprints to the hazard area.

        Returns
        -------
        None
        """
        gdf = self._get_fiat_gdf_full()

        crs = gdf.crs
        sfincs_extend = self.sfincs_overland_model.region
        sfincs_extend = sfincs_extend.to_crs(crs)

        # Clip the fiat region
        clipped_region = self.fiat_model.region.to_crs(crs).clip(sfincs_extend)
        self.fiat_model.geoms["region"] = clipped_region

        # Clip the exposure geometries
        gdf = self._clip_gdf(gdf, sfincs_extend, predicate="within")

        # Save exposure dataframe
        del gdf["geometry"]
        self.fiat_model.exposure.exposure_db = gdf.reset_index(drop=True)

        # Make
        self._delete_extra_geometries()

        # Clip the building footprints
        fieldname = "BF_FID"
        if clip_footprints and not self.fiat_model.building_footprint.empty:
            # Get buildings after filtering and their footprint id
            self.fiat_model.building_footprint = self.fiat_model.building_footprint[
                self.fiat_model.building_footprint[fieldname].isin(gdf[fieldname])
            ].reset_index(drop=True)

    @staticmethod
    def _clip_gdf(
        gdf1: gpd.GeoDataFrame, gdf2: gpd.GeoDataFrame, predicate: str = "within"
    ):
        gdf_new = gpd.sjoin(gdf1, gdf2, how="inner", predicate=predicate)
        gdf_new = gdf_new.drop(
            columns=[
                col
                for col in gdf_new.columns
                if col.endswith("_right") or (col in gdf2.columns and col != "geometry")
            ]
        )

        return gdf_new

    @staticmethod
    @debug_timer
    def spatial_join(
        objects: gpd.GeoDataFrame,
        layer: Union[str, gpd.GeoDataFrame],
        field_name: str,
        rename: Optional[str] = None,
        filter: Optional[bool] = False,
    ) -> tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]:
        """
        Perform a spatial join between two GeoDataFrames.

        Args:
            objects (gpd.GeoDataFrame): The GeoDataFrame representing the objects.
            layer (Union[str, gpd.GeoDataFrame]): The GeoDataFrame or file path of the layer to join with.
            field_name (str): The name of the field to use for the join.
            rename (Optional[str], optional): The new name to assign to the joined field. Defaults to None.

        Returns
        -------
            tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]: A tuple containing the joined GeoDataFrame and the layer GeoDataFrame.

        """
        # Read in layer and keep only column of interest
        if not isinstance(layer, gpd.GeoDataFrame):
            layer = gpd.read_file(layer)
        layer = layer[[field_name, "geometry"]]
        layer = layer.to_crs(objects.crs)
        if field_name in objects.columns:
            layer = layer.rename(columns={field_name: "layer_field"})
            field_name = "layer_field"
        # Spatial join of the layers
        objects_joined = objects.sjoin(layer, how="left", predicate="intersects")

        # Keep only the first intersection for each object
        objects_joined = (
            objects_joined.groupby(_FIAT_COLUMNS.object_id).first().reset_index()
        )

        # if needed filter out unused objects in the layer
        if filter:
            layer_inds = objects_joined["index_right"].dropna().unique()
            layer = layer.iloc[np.sort(layer_inds)].reset_index(drop=True)
        objects_joined = objects_joined[[_FIAT_COLUMNS.object_id, field_name]]
        # rename field if provided
        if rename:
            objects_joined = objects_joined.rename(columns={field_name: rename})
            layer = layer.rename(columns={field_name: rename})
        return objects_joined, layer

    def _get_fiat_building_index(self) -> list[int]:
        names = self.config.fiat_buildings_name
        if isinstance(names, str):
            names = [names]
        indices = [
            self.fiat_model.exposure.geom_names.index(name)
            for name in names
            if name in self.fiat_model.exposure.geom_names
        ]
        if indices:
            return indices
        raise ValueError(
            f"None of the specified building geometry names {names} found in FIAT model exposure geom_names."
        )

    def _get_fiat_road_index(self) -> int:
        return self.fiat_model.exposure.geom_names.index(self.config.fiat_roads_name)

    @debug_timer
    def _get_closest_station(self):
        # Get available stations from source
        obs_data = obs.source(self.config.tide_gauge.source)
        obs_data.get_active_stations()
        obs_stations = obs_data.gdf()
        # Calculate distance from SFINCS region to all available stations in degrees
        obs_stations["distance"] = obs_stations.distance(
            self.sfincs_overland_model.region.to_crs(4326).geometry.item()
        )
        # Get the closest station and its distance in meters
        closest_station = obs_stations[
            obs_stations["distance"] == obs_stations["distance"].min()
        ]
        distance = round(
            closest_station.to_crs(self.sfincs_overland_model.region.crs)
            .distance(self.sfincs_overland_model.region.geometry.item())
            .item(),
            0,
        )

        distance = us.UnitfulLength(value=distance, units=us.UnitTypesLength.meters)
        logger.info(
            f"The closest tide gauge from {self.config.tide_gauge.source} is located {distance.transform(self.unit_system.default_length_units)} from the SFINCS domain"
        )
        # Check if user provided max distance
        # TODO make sure units are explicit for max_distance
        if self.config.tide_gauge.max_distance is not None:
            units_new = self.config.tide_gauge.max_distance.units
            distance_new = us.UnitfulLength(
                value=distance.convert(units_new), units=units_new
            )
            if distance_new.value > self.config.tide_gauge.max_distance.value:
                logger.warning(
                    f"This distance is larger than the 'max_distance' value of {self.config.tide_gauge.max_distance.value} {units_new} provided in the config file. The station cannot be used."
                )
                return None

        # get station id
        station_id = closest_station["id"].item()

        return station_id

    @debug_timer
    def _get_station_metadata(self, station_id: str, ref: str = "MLLW"):
        """
        Find the closest tide gauge station to the SFINCS domain and retrieves its metadata.

        Args:
            station_id (str): The ID of the tide gauge station.
            ref (str, optional): The reference level for water level measurements. Defaults to "MLLW".

        Returns
        -------
            dict: A dictionary containing the metadata of the closest tide gauge station.
                The dictionary includes the following keys:
                - "id": The station ID.
                - "name": The station name.
                - "datum": The difference between the station's datum and the reference level.
                - "datum_name": The name of the datum used by the station.
                - "msl": The difference between the Mean Sea Level (MSL) and the reference level.
                - "reference": The reference level used for water level measurements.
                - "lon": The longitude of the station.
                - "lat": The latitude of the station.
        """
        # Get available stations from source
        obs_data = obs.source(self.config.tide_gauge.source)
        # read station metadata
        station_metadata = obs_data.get_meta_data(station_id)
        # TODO check if all stations can be used? Tidal attr?
        # Get water levels by using the ref provided
        datum_name = station_metadata["datums"]["OrthometricDatum"]
        datums = station_metadata["datums"]["datums"]
        names = [datum["name"] for datum in datums]

        ref_value = datums[names.index(ref)]["value"]

        meta = {
            "id": station_id,
            "name": station_metadata["name"],
            "datum": round(datums[names.index(datum_name)]["value"] - ref_value, 3),
            "datum_name": datum_name,
            "msl": round(datums[names.index("MSL")]["value"] - ref_value, 3),
            "mllw": round(datums[names.index("MLLW")]["value"] - ref_value, 3),
            "mhhw": round(datums[names.index("MHHW")]["value"] - ref_value, 3),
            "reference": ref,
            "units": station_metadata["datums"]["units"],
            "lon": station_metadata["lng"],
            "lat": station_metadata["lat"],
        }

        logger.info(
            f"The tide gauge station '{station_metadata['name']}' from {self.config.tide_gauge.source} will be used to download nearshore historical water level time-series."
        )

        logger.info(
            f"The station metadata will be used to fill in the water_level attribute in the site.toml. The reference level will be '{ref}'."
        )

        return meta

    def _get_bin_colors(self):
        """
        Retrieve the bin colors from the bin_colors.toml file.

        Returns
        -------
            dict: A dictionary containing the bin colors.
        """
        templates_path = Path(__file__).parent.resolve().joinpath("templates")
        with open(
            templates_path.joinpath("output_layers", "bin_colors.toml"), "rb"
        ) as f:
            bin_colors = tomli.load(f)
        return bin_colors

    def _delete_extra_geometries(self) -> None:
        """
        Remove extra geometries from the exposure_geoms list that do not have a corresponding object_id in the exposure_db DataFrame.

        Returns
        -------
            None
        """
        # Make sure only csv objects have geometries
        for i, geoms in enumerate(self.fiat_model.exposure.exposure_geoms):
            keep = geoms[_FIAT_COLUMNS.object_id].isin(
                self.fiat_model.exposure.exposure_db[_FIAT_COLUMNS.object_id]
            )
            geoms = geoms[keep].reset_index(drop=True)
            self.fiat_model.exposure.exposure_geoms[i] = geoms

    def _get_fiat_gdf_full(self) -> gpd.GeoDataFrame:
        """
        Get the full GeoDataFrame of the Fiat model.

        Returns
        -------
            gpd.GeoDataFrame: The full GeoDataFrame of the Fiat model.
        """
        gdf = self.fiat_model.exposure.get_full_gdf(
            self.fiat_model.exposure.exposure_db
        )
        # Keep only unique "object_id" rows, keeping the first occurrence
        gdf = gdf.drop_duplicates(
            subset=_FIAT_COLUMNS.object_id, keep="first"
        ).reset_index(drop=True)

        return gdf


def create_database(config: Union[str, Path, ConfigModel], overwrite=False) -> None:
    """Create a new database from a configuration file or ConfigModel.

    Parameters
    ----------
    config : str, Path, or ConfigModel
        The path to the configuration file (as a string or Path) or a ConfigModel instance.
    overwrite : bool, default False
        Whether to overwrite the existing database if it exists.
    """
    if isinstance(config, (str, Path)):
        config = ConfigModel.read(config)

    DatabaseBuilder(config=config).build(overwrite)


def main():
    while True:
        config_path = Path(
            input(
                "Please provide the path to the database creation configuration toml: \n"
            )
        )
        print(
            "Please select the log verbosity level for the database creation process.\n"
            "From most verbose to least verbose: `DEBUG`, `INFO`, `WARNING`.'n"
        )
        log_level = input("Enter log level: ")
        match log_level:
            case "DEBUG":
                level = logging.DEBUG
            case "INFO":
                level = logging.INFO
            case "WARNING":
                level = logging.WARNING
            case _:
                print(
                    f"Log level `{log_level}` not recognized. Defaulting to INFO. Please choose from: `DEBUG`, `INFO`, `WARNING`."
                )
                log_level = "INFO"
                level = logging.INFO

        FloodAdaptLogging(level=level)

        try:
            config = ConfigModel.read(config_path)
            dbs = DatabaseBuilder(config)
            dbs.build()
        except Exception as e:
            print(e)
        quit = input("Do you want to quit? (y/n)")
        if quit == "y":
            exit()


if __name__ == "__main__":
    main()



import os
from pathlib import Path
from typing import Annotated

import pandas as pd

from flood_adapt.misc.utils import (
    copy_file_to_output_dir,
    validate_file_extension,
)
from flood_adapt.objects.forcing import unit_system as us
from flood_adapt.objects.forcing.forcing import (
    ForcingSource,
    IDischarge,
)
from flood_adapt.objects.forcing.time_frame import TimeFrame
from flood_adapt.objects.forcing.timeseries import (
    CSVTimeseries,
    SyntheticTimeseries,
    TimeseriesFactory,
)


class DischargeConstant(IDischarge):
    source: ForcingSource = ForcingSource.CONSTANT

    discharge: us.UnitfulDischarge

    def to_dataframe(self, time_frame: TimeFrame) -> pd.DataFrame:
        time = pd.date_range(
            start=time_frame.start_time,
            end=time_frame.end_time,
            freq=time_frame.time_step,
            name="time",
        )
        data = [self.discharge.value for _ in range(len(time))]
        return pd.DataFrame(index=time, data=data, columns=[self.river.name])


class DischargeSynthetic(IDischarge):
    source: ForcingSource = ForcingSource.SYNTHETIC

    timeseries: SyntheticTimeseries

    def to_dataframe(self, time_frame: TimeFrame) -> pd.DataFrame:
        df = TimeseriesFactory.from_object(self.timeseries).to_dataframe(
            time_frame=time_frame
        )
        df.columns = [self.river.name]
        return df


class DischargeCSV(IDischarge):
    source: ForcingSource = ForcingSource.CSV

    path: Annotated[Path, validate_file_extension([".csv"])]

    units: us.UnitTypesDischarge = us.UnitTypesDischarge.cms

    def to_dataframe(self, time_frame: TimeFrame) -> pd.DataFrame:
        return CSVTimeseries.load_file(
            path=self.path, units=us.UnitfulDischarge(value=0, units=self.units)
        ).to_dataframe(time_frame=time_frame)

    def save_additional(self, output_dir: Path | str | os.PathLike) -> None:
        self.path = copy_file_to_output_dir(self.path, Path(output_dir))



from abc import ABC, abstractmethod
from pathlib import Path

import pandas as pd

from flood_adapt.objects.scenarios.scenarios import Scenario


class IOffshoreSfincsHandler(ABC):
    template_path: Path

    @abstractmethod
    def get_resulting_waterlevels(self, scenario: Scenario) -> pd.DataFrame: ...

    @staticmethod
    def requires_offshore_run(scenario: Scenario) -> bool: ...

    @abstractmethod
    def run_offshore(self, scenario: Scenario): ...



# has to be here at the start to avoid circular imports
__version__ = "1.0.5"

from flood_adapt import adapter, database_builder, dbs_classes, objects
from flood_adapt.config.config import Settings
from flood_adapt.config.site import Site
from flood_adapt.flood_adapt import FloodAdapt
from flood_adapt.misc.exceptions import DatabaseError, FloodAdaptError
from flood_adapt.misc.log import FloodAdaptLogging
from flood_adapt.objects.forcing import unit_system

__all__ = [
    "FloodAdapt",
    "Site",
    "Settings",
    "FloodAdaptLogging",
    "unit_system",
    "objects",
    "dbs_classes",
    "adapter",
    "database_builder",
    "FloodAdaptError",
    "DatabaseError",
]

FloodAdaptLogging()  # Initialize logging once for the entire package



from datetime import datetime, timedelta

from pydantic import (
    BaseModel,
    field_validator,
    model_validator,
)

from flood_adapt.objects.forcing import unit_system as us

REFERENCE_TIME = datetime(year=2021, month=1, day=1, hour=0, minute=0, second=0)


class TimeFrame(BaseModel):
    """
    Class representing a time frame for a simulation.

    Attributes
    ----------
    start_time : datetime
        The start time of the simulation.
    end_time : datetime
        The end time of the simulation.
    time_step : timedelta
        The time step of the simulation. Default is calculated as 1/1000 of the duration.
    """

    start_time: datetime = REFERENCE_TIME
    end_time: datetime = REFERENCE_TIME + timedelta(days=1)

    @property
    def time_step(self) -> timedelta:
        return self._time_step

    @field_validator("start_time", "end_time", mode="before")
    @classmethod
    def try_parse_datetime(cls, value: str | datetime) -> datetime:
        SUPPORTED_DATETIME_FORMATS = [
            "%Y%m%d %H%M%S",
            "%Y-%m-%d %H:%M:%S",
            "%Y-%m-%d %H:%M:%S.%f",
            "%Y-%m-%d %H:%M:%S.%f%z",
        ]
        if not isinstance(value, datetime):
            for fmt in SUPPORTED_DATETIME_FORMATS:
                try:
                    value = datetime.strptime(value, fmt)
                    break
                except Exception:
                    pass

        if not isinstance(value, datetime):
            raise ValueError(
                f"Could not parse start time: {value}. Supported formats are {', '.join(SUPPORTED_DATETIME_FORMATS)}"
            )
        return value

    @model_validator(mode="after")
    def start_time_before_end_time(self):
        if self.start_time >= self.end_time:
            raise ValueError(
                f"Start time: {self.start_time} must be before end time: {self.end_time}"
            )
        return self

    @model_validator(mode="after")
    def dynamic_timestep(self):
        num_intervals = 1000
        time_span = (self.end_time - self.start_time).total_seconds()

        # Round the time step to the second for simplicity
        self._time_step = timedelta(
            seconds=int(timedelta(seconds=time_span / num_intervals).total_seconds())
        )
        return self

    @property
    def duration(self) -> timedelta:
        return self.end_time - self.start_time

    def duration_as_unitfultime(
        self, unit: us.UnitTypesTime = us.UnitTypesTime.hours
    ) -> us.UnitfulTime:
        return us.UnitfulTime(
            value=(self.duration).total_seconds(),
            units=us.UnitTypesTime.seconds,
        ).transform(unit)

    def __str__(self) -> str:
        return f"{self.start_time} - {self.end_time}"
